"""Generate rule documentation."""

import abc
import inspect
import textwrap

from dbt_score.config import Config
from dbt_score.rule import Rule
from dbt_score.rule_registry import RuleRegistry


class Formatter(abc.ABC):
    """A rule catalog formatter."""

    @staticmethod
    @abc.abstractmethod
    def header(title: str | None) -> str | None:
        """Return text at the top."""
        raise NotImplementedError

    @staticmethod
    @abc.abstractmethod
    def footer() -> str | None:
        """Return text at the bottom."""
        raise NotImplementedError

    @staticmethod
    @abc.abstractmethod
    def format_rule(rule: Rule) -> str | None:
        """Return text for a given rule."""
        raise NotImplementedError


class PlainTextFormatter(Formatter):
    """A terminal formatter."""

    @staticmethod
    def header(title: str | None) -> str | None:
        """Return text at the top."""
        return None

    @staticmethod
    def footer() -> str | None:
        """Return text at the bottom."""
        return None

    @staticmethod
    def bold(text: str) -> str:
        """Return text in bold."""
        return f"\033[1m{text}\033[0m"

    @staticmethod
    def format_rule(rule: Rule) -> str | None:
        """Return text for a given rule."""
        return f"{PlainTextFormatter.bold(rule.source())}:\n    {rule.description}\n"


class MarkdownFormatter(Formatter):
    """A formatter for MkDocs (Markdown)."""

    @staticmethod
    def header(title: str | None) -> str | None:
        """Return text at the top."""
        if title:
            return f"# {title}\n"
        return "# Rules\n"

    @staticmethod
    def footer() -> str | None:
        """Return text at the bottom."""
        return None

    @staticmethod
    def format_rule(rule: Rule) -> str | None:
        """Return text for a given rule."""
        rule_name = rule.source().split(".")[-1]

        rule_description = rule.description

        if hasattr(rule, "_orig_evaluate"):
            rule_source_code = inspect.getsource(rule._orig_evaluate)
        else:
            rule_source_code = inspect.getsource(rule.__class__)
        rule_source_code = textwrap.indent(rule_source_code, " " * 4)

        rule_configuration = f'[tool.dbt-score.rules."{rule.source()}"]\n'
        rule_configuration += f"severity = {rule.severity.value}"
        for config_key, config_default_value in rule.default_config.items():
            rule_configuration += f"\n{config_key} = {config_default_value}"

        template = f"""## `{rule_name}`

{rule_description}

??? quote  "Source code"
    ```python
{rule_source_code}
    ```

### Default configuration

```toml title="pyproject.toml"
{rule_configuration}
```
"""
        return template


def display_catalog(config: Config, title: str, format: str) -> None:
    """Print rules catalog."""
    formatter: Formatter
    if format == "terminal":
        formatter = PlainTextFormatter()
    elif format == "markdown":
        formatter = MarkdownFormatter()
    else:
        raise Exception(f"Format {format} is not valid.")

    rule_registry = RuleRegistry(config)
    rule_registry.load_all()

    header = formatter.header(title)
    if header:
        print(header)
    for rule in rule_registry.rules.values():
        rule_doc = formatter.format_rule(rule)
        if rule_doc:
            print(rule_doc)
    footer = formatter.footer()
    if footer:
        print(footer)



"""Test rule."""

import pytest
from dbt_score import Model, Rule, RuleViolation, Seed, Severity, Snapshot, Source, rule
from dbt_score.rule_filter import RuleFilter, rule_filter


def test_rule_decorator_and_class(
    decorator_rule,
    decorator_rule_no_parens,
    decorator_rule_args,
    class_rule,
    model1,
    model2,
):
    """Test rule creation with the rule decorator and class."""
    decorator_rule_instance = decorator_rule()
    decorator_rule_no_parens_instance = decorator_rule_no_parens()
    decorator_rule_args_instance = decorator_rule_args()
    class_rule_instance = class_rule()

    def assertions(rule_instance):
        assert isinstance(rule_instance, Rule)
        assert rule_instance.severity == Severity.MEDIUM
        assert rule_instance.description == "Description of the rule."
        assert rule_instance.evaluate(model1) == RuleViolation(
            message="Model1 is a violation."
        )
        assert rule_instance.evaluate(model2) is None

    assertions(decorator_rule_instance)
    assertions(decorator_rule_no_parens_instance)
    assertions(decorator_rule_args_instance)
    assertions(class_rule_instance)


def test_missing_description_rule_decorator():
    """Test missing description in rule decorator."""
    with pytest.raises(AttributeError):

        @rule()
        def example_rule(model: Model) -> RuleViolation | None:
            return None


def test_missing_description_rule_class():
    """Test missing description in rule class."""
    with pytest.raises(AttributeError):

        class BadRule(Rule):
            """Bad example rule."""

            def evaluate(self, model: Model) -> RuleViolation | None:  # type: ignore[override]
                """Evaluate model."""
                return None


def test_missing_evaluate_rule_class(model1):
    """Test missing evaluate implementation in rule class."""
    with pytest.raises(TypeError):

        class BadRule(Rule):
            """Bad example rule."""

            description = "Description of the rule."


@pytest.mark.parametrize(
    "rule_fixture,expected_type",
    [
        ("decorator_rule", Model),
        ("decorator_rule_no_parens", Model),
        ("decorator_rule_args", Model),
        ("class_rule", Model),
        ("decorator_rule_source", Source),
        ("decorator_rule_no_parens_source", Source),
        ("decorator_rule_args_source", Source),
        ("class_rule_source", Source),
        ("decorator_rule_snapshot", Snapshot),
        ("decorator_rule_no_parens_snapshot", Snapshot),
        ("decorator_rule_args_snapshot", Snapshot),
        ("class_rule_snapshot", Snapshot),
        ("decorator_rule_seed", Seed),
        ("decorator_rule_no_parens_seed", Seed),
        ("decorator_rule_args_seed", Seed),
        ("class_rule_seed", Seed),
    ],
)
def test_rule_introspects_its_resource_type(request, rule_fixture, expected_type):
    """Test that each rule is aware of the resource-type it is evaluated against."""
    rule = request.getfixturevalue(rule_fixture)
    assert rule().resource_type is expected_type


class TestRuleFilterValidation:
    """Tests that a rule filter matches resource-type to the rule it's attached to."""

    @pytest.fixture
    def source_filter_no_parens(self):
        """Example source filter with bare decorator."""

        @rule_filter
        def source_filter(source: Source) -> bool:
            """Description."""
            return False

        return source_filter()

    @pytest.fixture
    def source_filter_parens(self):
        """Example source filter with decorator and parens."""

        @rule_filter()
        def source_filter(source: Source) -> bool:
            """Description."""
            return False

        return source_filter()

    @pytest.fixture
    def source_filter_class(self):
        """Example class-based source filter."""

        class SourceFilter(RuleFilter):
            description = "Description"

            def evaluate(self, source: Source) -> bool:  # type: ignore[override]
                return False

        return SourceFilter

    @pytest.mark.parametrize(
        "rule_filter_fixture",
        ["source_filter_no_parens", "source_filter_parens", "source_filter_class"],
    )
    def test_rule_filter_must_match_resource_type_as_rule(
        self, request, rule_filter_fixture
    ):
        """Tests that rules can't be created with filters of incorrect resource-type."""
        rule_filter = request.getfixturevalue(rule_filter_fixture)

        with pytest.raises(TypeError) as excinfo:

            @rule(rule_filters={rule_filter})
            def model_always_passes(model: Model) -> RuleViolation | None:
                """Description."""
                pass

        assert "Mismatched resource_type on filter" in str(excinfo.value)
        assert "Expected Model, but got Source" in str(excinfo.value)

        with pytest.raises(TypeError):

            class ModelAlwaysPasses(Rule):
                description = "Description."
                rule_filters = frozenset([rule_filter])

                def evaluate(self, model: Model) -> RuleViolation | None:  # type: ignore[override]
                    pass

        assert "Mismatched resource_type on filter" in str(excinfo.value)
        assert "Expected Model, but got Source" in str(excinfo.value)



"""Vendored utility functions from https://github.com/more-itertools/more-itertools."""
from typing import (
    Callable,
    Iterable,
    Optional,
    TypeVar,
    overload,
)

_T = TypeVar("_T")
_U = TypeVar("_U")


@overload
def first_true(
    iterable: Iterable[_T], *, pred: Callable[[_T], object] | None = ...
) -> _T | None:
    ...


@overload
def first_true(
    iterable: Iterable[_T],
    default: _U,
    pred: Callable[[_T], object] | None = ...,
) -> _T | _U:
    ...


def first_true(
    iterable: Iterable[_T],
    default: Optional[_U] = None,
    pred: Optional[Callable[[_T], object]] = None,
) -> _T | _U | None:
    """Returns the first true value in the iterable.

    If no true value is found, returns *default*

    If *pred* is not None, returns the first item for which
    ``pred(item) == True`` .

        >>> first_true(range(10))
        1
        >>> first_true(range(10), pred=lambda x: x > 5)
        6
        >>> first_true(range(10), default='missing', pred=lambda x: x > 9)
        'missing'

    """
    return next(filter(pred, iterable), default)



"""Rule registry.

This module implements rule discovery.
"""

import importlib
import logging
import os
import pkgutil
import sys
from typing import Iterator, Type

from dbt_score.config import Config
from dbt_score.exceptions import DuplicatedRuleException
from dbt_score.rule import Rule, RuleConfig
from dbt_score.rule_filter import RuleFilter

logger = logging.getLogger(__name__)


class RuleRegistry:
    """A container for configured rules."""

    def __init__(self, config: Config) -> None:
        """Instantiate a rule registry."""
        self.config = config
        self._rules: dict[str, Rule] = {}
        self._rule_filters: dict[str, RuleFilter] = {}

    @property
    def rules(self) -> dict[str, Rule]:
        """Get all rules."""
        return self._rules

    @property
    def rule_filters(self) -> dict[str, RuleFilter]:
        """Get all filters."""
        return self._rule_filters

    def _walk_packages(self, namespace_name: str) -> Iterator[str]:
        """Walk packages and sub-packages recursively."""
        try:
            namespace = importlib.import_module(namespace_name)
        except ImportError:  # no custom rule in Python path
            if namespace_name != "dbt_score_rules":
                logger.warning(f"Can't import {namespace_name}.")
            return

        if not hasattr(namespace, "__path__"):
            # When called with a leaf, i.e. a module, don't attempt to iterate
            yield namespace_name
            return

        for package in pkgutil.walk_packages(
            namespace.__path__, namespace.__name__ + "."
        ):
            yield package.name

    def _load(self, namespace_name: str) -> None:
        """Load rules and filters found in a given namespace."""
        for module_name in self._walk_packages(namespace_name):
            module = importlib.import_module(module_name)
            for obj_name in dir(module):
                obj = module.__dict__[obj_name]
                # Skip adding objects imported from other modules
                if type(obj) is type and module.__name__ != obj.__module__:
                    continue
                if type(obj) is type and issubclass(obj, Rule) and obj is not Rule:
                    self._add_rule(obj)
                if (
                    type(obj) is type
                    and issubclass(obj, RuleFilter)
                    and obj is not RuleFilter
                ):
                    self._add_filter(obj)

    def _add_rule(self, rule: Type[Rule]) -> None:
        """Initialize and add a rule."""
        rule_name = rule.source()
        if rule_name in self._rules:
            raise DuplicatedRuleException(rule_name)
        if rule_name not in self.config.disabled_rules:
            rule_config = self.config.rules_config.get(rule_name, RuleConfig())
            self._rules[rule_name] = rule(rule_config=rule_config)

    def _add_filter(self, rule_filter: Type[RuleFilter]) -> None:
        """Initialize and add a filter."""
        filter_name = rule_filter.source()
        if filter_name in self._rule_filters:
            raise DuplicatedRuleException(filter_name)
        self._rule_filters[filter_name] = rule_filter()

    def load_all(self) -> None:
        """Load all rules, core and third-party."""
        # Add cwd to Python path
        old_sys_path = sys.path  # Save original values
        if self.config.inject_cwd_in_python_path and os.getcwd() not in sys.path:
            sys.path.append(os.getcwd())

        for namespace in self.config.rule_namespaces:
            self._load(namespace)

        # Restore original values
        sys.path = old_sys_path

        self._load_filters_into_rules()

    def _load_filters_into_rules(self) -> None:
        """Loads RuleFilters into Rule objects.

        If the config of the rule has filter names in the `rule_filter_names` key,
        load those filters from the rule registry into the actual `rule_filters` field.
        Configuration overwrites any pre-existing filters.
        """
        for rule in self._rules.values():
            filter_names: list[str] = rule.rule_filter_names or []
            if len(filter_names) > 0:
                rule.set_filters(
                    rule_filter
                    for name, rule_filter in self.rule_filters.items()
                    if name in filter_names
                )



"""Rule definitions."""

import inspect
import typing
from dataclasses import dataclass, field
from enum import Enum
from typing import (
    Any,
    Callable,
    Iterable,
    Type,
    TypeAlias,
    cast,
    overload,
)

from dbt_score.models import Evaluable, Model, Seed, Snapshot, Source
from dbt_score.more_itertools import first_true
from dbt_score.rule_filter import RuleFilter


class Severity(Enum):
    """The severity/weight of a rule."""

    LOW = 1
    MEDIUM = 2
    HIGH = 3
    CRITICAL = 4


@dataclass
class RuleConfig:
    """Configuration for a rule."""

    severity: Severity | None = None
    config: dict[str, Any] = field(default_factory=dict)
    rule_filter_names: list[str] = field(default_factory=list)

    @staticmethod
    def from_dict(rule_config: dict[str, Any]) -> "RuleConfig":
        """Create a RuleConfig from a dictionary."""
        config = rule_config.copy()
        severity = (
            Severity(config.pop("severity", None))
            if "severity" in rule_config
            else None
        )
        filter_names = (
            config.pop("rule_filter_names", None)
            if "rule_filter_names" in rule_config
            else []
        )

        return RuleConfig(
            severity=severity, config=config, rule_filter_names=filter_names
        )


@dataclass
class RuleViolation:
    """The violation of a rule."""

    message: str | None = None


ModelRuleEvaluationType: TypeAlias = Callable[[Model], RuleViolation | None]
SourceRuleEvaluationType: TypeAlias = Callable[[Source], RuleViolation | None]
SnapshotRuleEvaluationType: TypeAlias = Callable[[Snapshot], RuleViolation | None]
SeedRuleEvaluationType: TypeAlias = Callable[[Seed], RuleViolation | None]
RuleEvaluationType: TypeAlias = (
    ModelRuleEvaluationType
    | SourceRuleEvaluationType
    | SnapshotRuleEvaluationType
    | SeedRuleEvaluationType
)


class Rule:
    """The rule base class."""

    description: str
    severity: Severity = Severity.MEDIUM
    rule_filter_names: list[str]
    rule_filters: frozenset[RuleFilter] = frozenset()
    default_config: typing.ClassVar[dict[str, Any]] = {}
    resource_type: typing.ClassVar[type[Evaluable]]

    def __init__(self, rule_config: RuleConfig | None = None) -> None:
        """Initialize the rule."""
        self.config: dict[str, Any] = {}
        if rule_config:
            self.process_config(rule_config)

    def __init_subclass__(cls, **kwargs) -> None:  # type: ignore
        """Initializes the subclass."""
        super().__init_subclass__(**kwargs)
        if not hasattr(cls, "description"):
            raise AttributeError("Subclass must define class attribute `description`.")

        cls.resource_type = cls._introspect_resource_type()

        cls._validate_rule_filters()

    @classmethod
    def _validate_rule_filters(cls) -> None:
        for rule_filter in cls.rule_filters:
            if rule_filter.resource_type != cls.resource_type:
                raise TypeError(
                    f"Mismatched resource_type on filter "
                    f"{rule_filter.__class__.__name__}. "
                    f"Expected {cls.resource_type.__name__}, "
                    f"but got {rule_filter.resource_type.__name__}."
                )

    @classmethod
    def _introspect_resource_type(cls) -> Type[Evaluable]:
        evaluate_func = getattr(cls, "_orig_evaluate", cls.evaluate)

        sig = inspect.signature(evaluate_func)
        resource_type_argument = first_true(
            sig.parameters.values(),
            pred=lambda arg: arg.annotation in typing.get_args(Evaluable),
        )

        if not resource_type_argument:
            raise TypeError(
                "Subclass must implement method `evaluate` with an "
                "annotated Model or Source argument."
            )

        resource_type = cast(type[Evaluable], resource_type_argument.annotation)
        return resource_type

    def process_config(self, rule_config: RuleConfig) -> None:
        """Process the rule config."""
        config = self.default_config.copy()

        # Overwrite default rule configuration
        for k, v in rule_config.config.items():
            if k in self.default_config:
                config[k] = v
            else:
                raise AttributeError(
                    f"Unknown rule parameter: {k} for rule {self.source()}."
                )

        self.set_severity(
            rule_config.severity
        ) if rule_config.severity else rule_config.severity
        self.rule_filter_names = rule_config.rule_filter_names
        self.config = config

    def evaluate(self, evaluable: Evaluable) -> RuleViolation | None:
        """Evaluates the rule."""
        raise NotImplementedError("Subclass must implement method `evaluate`.")

    @classmethod
    def should_evaluate(cls, evaluable: Evaluable) -> bool:
        """Checks whether the rule should be applied against the evaluable.

        The evaluable must satisfy the following criteria:
            - all filters in the rule allow evaluation
            - the rule and evaluable have matching resource_types
        """
        resource_types_match = cls.resource_type is type(evaluable)

        if cls.rule_filters:
            return (
                all(f.evaluate(evaluable) for f in cls.rule_filters)
                and resource_types_match
            )
        return resource_types_match

    @classmethod
    def set_severity(cls, severity: Severity) -> None:
        """Set the severity of the rule."""
        cls.severity = severity

    @classmethod
    def set_filters(cls, rule_filters: Iterable[RuleFilter]) -> None:
        """Set the filters of the rule."""
        cls.rule_filters = frozenset(rule_filters)

    @classmethod
    def source(cls) -> str:
        """Return the source of the rule, i.e. a fully qualified name."""
        return f"{cls.__module__}.{cls.__name__}"

    def __hash__(self) -> int:
        """Compute a unique hash for a rule."""
        return hash(self.source())


# Use @overload to have proper typing for both @rule and @rule(...)
# https://mypy.readthedocs.io/en/stable/generics.html#decorator-factories


@overload
def rule(__func: ModelRuleEvaluationType) -> Type[Rule]:
    ...


@overload
def rule(__func: SourceRuleEvaluationType) -> Type[Rule]:
    ...


@overload
def rule(__func: SnapshotRuleEvaluationType) -> Type[Rule]:
    ...


@overload
def rule(__func: SeedRuleEvaluationType) -> Type[Rule]:
    ...


@overload
def rule(
    *,
    description: str | None = None,
    severity: Severity = Severity.MEDIUM,
    rule_filters: set[RuleFilter] | None = None,
) -> Callable[[RuleEvaluationType], Type[Rule]]:
    ...


def rule(
    __func: RuleEvaluationType | None = None,
    *,
    description: str | None = None,
    severity: Severity = Severity.MEDIUM,
    rule_filters: set[RuleFilter] | None = None,
) -> Type[Rule] | Callable[[RuleEvaluationType], Type[Rule]]:
    """Rule decorator.

    The rule decorator creates a rule class (subclass of Rule) and returns it.

    Using arguments or not are both supported:
    - ``@rule``
    - ``@rule(description="...")``

    Args:
        __func: The rule evaluation function being decorated.
        description: The description of the rule.
        severity: The severity of the rule.
        rule_filters: Set of RuleFilter that filters the items that the rule applies to.
    """

    def decorator_rule(func: RuleEvaluationType) -> Type[Rule]:
        """Decorator function."""
        if func.__doc__ is None and description is None:
            raise AttributeError("Rule must define `description` or `func.__doc__`.")

        # Get description parameter, otherwise use the docstring
        rule_description = description or (
            func.__doc__.split("\n")[0] if func.__doc__ else None
        )

        def wrapped_func(self: Rule, *args: Any, **kwargs: Any) -> RuleViolation | None:
            """Wrap func to add `self`."""
            return func(*args, **kwargs)

        # Get default parameters from the rule definition
        default_config = {
            key: val.default
            for key, val in inspect.signature(func).parameters.items()
            if val.default != inspect.Parameter.empty
        }

        # Create the rule class inheriting from Rule
        rule_class = type(
            func.__name__,
            (Rule,),
            {
                "description": rule_description,
                "severity": severity,
                "rule_filters": rule_filters or frozenset(),
                "default_config": default_config,
                "evaluate": wrapped_func,
                # Save provided evaluate function
                "_orig_evaluate": func,
                # Forward origin of the decorated function
                "__qualname__": func.__qualname__,  # https://peps.python.org/pep-3155/
                "__module__": func.__module__,
            },
        )

        return rule_class

    if __func is not None:
        # The syntax @rule is used
        return decorator_rule(__func)
    else:
        # The syntax @rule(...) is used
        return decorator_rule



"""Tests for vendored more_itertools functions."""
from dbt_score.more_itertools import first_true


class TestFirstTrue:
    """Tests for `first_true`."""

    def test_something_true(self):
        """Test with no keyword arguments."""
        assert first_true(range(10), 1)

    def test_nothing_true(self):
        """Test default return value."""
        assert first_true([0, 0, 0]) is None

    def test_default(self):
        """Test with a default keyword."""
        assert first_true([0, 0, 0], default="!") == "!"

    def test_pred(self):
        """Test with a custom predicate."""
        assert first_true([2, 4, 6], pred=lambda x: x % 3 == 0) == 6



"""Unit tests for the human readable formatter."""

from textwrap import dedent

import pytest
from dbt_score.evaluation import EvaluableResultsType
from dbt_score.formatters.human_readable_formatter import HumanReadableFormatter
from dbt_score.rule import RuleViolation
from dbt_score.scoring import Score


def test_human_readable_formatter_model_with_defaults(
    capsys,
    default_config,
    manifest_loader,
    model1,
    rule_severity_low,
    rule_severity_medium,
    rule_severity_critical,
):
    """Ensure the formatter has the correct output after model evaluation."""
    formatter = HumanReadableFormatter(
        manifest_loader=manifest_loader, config=default_config
    )
    results: EvaluableResultsType = {
        rule_severity_low: None,
        rule_severity_medium: Exception("Oh noes"),
        rule_severity_critical: RuleViolation("Error"),
    }
    formatter.evaluable_evaluated(model1, results, Score(10.0, "🥇"))
    stdout = capsys.readouterr().out
    expected = """\
    🥇 \x1b[1mModel: model1\x1b[0m (score: 10.0)
        \x1b[1;31mERR \x1b[0m tests.conftest.rule_severity_medium: Oh noes
        \x1b[1;33mWARN\x1b[0m (critical) tests.conftest.rule_severity_critical: Error

    """
    assert stdout == dedent(expected)


@pytest.mark.parametrize(
    "show,expected",
    [
        (
            "all",
            """\
    🥇 \x1b[1mModel: model1\x1b[0m (score: 10.0)
        \x1b[1;32mOK  \x1b[0m tests.conftest.rule_severity_low
        \x1b[1;31mERR \x1b[0m tests.conftest.rule_severity_medium: Oh noes
        \x1b[1;33mWARN\x1b[0m (critical) tests.conftest.rule_severity_critical: Error

    """,
        ),
        (
            "failing-rules",
            """\
    🥇 \x1b[1mModel: model1\x1b[0m (score: 10.0)
        \x1b[1;31mERR \x1b[0m tests.conftest.rule_severity_medium: Oh noes
        \x1b[1;33mWARN\x1b[0m (critical) tests.conftest.rule_severity_critical: Error

    """,
        ),
        (
            "failing-items",
            "",
        ),
    ],
)
def test_human_readable_formatter_model_show_parameter(
    capsys,
    default_config,
    manifest_loader,
    model1,
    rule_severity_low,
    rule_severity_medium,
    rule_severity_critical,
    show,
    expected,
):
    """Ensure the formatter has the correct output after model evaluation."""
    default_config.overload({"show": show})
    formatter = HumanReadableFormatter(
        manifest_loader=manifest_loader, config=default_config
    )
    results: EvaluableResultsType = {
        rule_severity_low: None,
        rule_severity_medium: Exception("Oh noes"),
        rule_severity_critical: RuleViolation("Error"),
    }
    formatter.evaluable_evaluated(model1, results, Score(10.0, "🥇"))
    stdout = capsys.readouterr().out
    assert stdout == dedent(expected)


def test_human_readable_formatter_project(capsys, default_config, manifest_loader):
    """Ensure the formatter has the correct output after project evaluation."""
    formatter = HumanReadableFormatter(
        manifest_loader=manifest_loader, config=default_config
    )
    formatter.project_evaluated(Score(10.0, "🥇"))
    stdout = capsys.readouterr().out
    assert stdout == "Project score: \x1b[1m10.0\x1b[0m 🥇\n"


def test_human_readable_formatter_near_perfect_model_score(
    capsys,
    default_config,
    manifest_loader,
    model1,
    rule_severity_low,
    rule_severity_medium,
    rule_severity_critical,
):
    """Ensure the formatter has the correct output after model evaluation."""
    default_config.overload({"show": "all"})
    formatter = HumanReadableFormatter(
        manifest_loader=manifest_loader, config=default_config
    )
    results: EvaluableResultsType = {
        rule_severity_low: None,
        rule_severity_medium: Exception("Oh noes"),
        rule_severity_critical: RuleViolation("Error"),
    }
    formatter.evaluable_evaluated(model1, results, Score(9.99, "🥈"))
    stdout = capsys.readouterr().out

    expected = """\
    🥈 \x1b[1mModel: model1\x1b[0m (score: 9.9)
        \x1b[1;32mOK  \x1b[0m tests.conftest.rule_severity_low
        \x1b[1;31mERR \x1b[0m tests.conftest.rule_severity_medium: Oh noes
        \x1b[1;33mWARN\x1b[0m (critical) tests.conftest.rule_severity_critical: Error

    """
    assert stdout == dedent(expected)


def test_human_readable_formatter_near_perfect_project_score(
    capsys, default_config, manifest_loader
):
    """Ensure the formatter has the correct output after project evaluation."""
    formatter = HumanReadableFormatter(
        manifest_loader=manifest_loader, config=default_config
    )
    formatter.project_evaluated(Score(9.99, "🥈"))
    stdout = capsys.readouterr().out
    assert stdout == "Project score: \x1b[1m9.9\x1b[0m 🥈\n"


def test_human_readable_formatter_low_evaluable_score(
    capsys,
    default_config,
    manifest_loader,
    model1,
    source1,
    rule_severity_critical,
):
    """Ensure the formatter has the correct output when a model has a low score."""
    formatter = HumanReadableFormatter(
        manifest_loader=manifest_loader, config=default_config
    )
    results: EvaluableResultsType = {
        rule_severity_critical: RuleViolation("Error"),
    }
    formatter.evaluable_evaluated(model1, results, Score(0.0, "🚧"))
    formatter.evaluable_evaluated(source1, results, Score(0.0, "🚧"))
    formatter.project_evaluated(Score(0.0, "🚧"))
    stdout = capsys.readouterr().out

    expected = """\
    🚧 \x1b[1mModel: model1\x1b[0m (score: 0.0)
        \x1b[1;33mWARN\x1b[0m (critical) tests.conftest.rule_severity_critical: Error

    🚧 \x1b[1mSource: my_source.table1\x1b[0m (score: 0.0)
        \x1b[1;33mWARN\x1b[0m (critical) tests.conftest.rule_severity_critical: Error

    Project score: \x1b[1m0.0\x1b[0m 🚧

    Error: evaluable score too low, fail_any_item_under = 5.0
    Model model1 scored 0.0
    Source my_source.table1 scored 0.0
    """
    assert stdout == dedent(expected)


def test_human_readable_formatter_low_project_score_high_model_score(
    capsys,
    default_config,
    manifest_loader,
    model1,
    rule_severity_critical,
):
    """Ensure the formatter has the correct output when the projet has a low score.

    If model itself has a high project score then we need to pass `show_all` flag
    to make it visible.
    """
    default_config.overload({"show_all": True})
    formatter = HumanReadableFormatter(
        manifest_loader=manifest_loader, config=default_config
    )
    results: EvaluableResultsType = {
        rule_severity_critical: RuleViolation("Error"),
    }
    formatter.evaluable_evaluated(model1, results, Score(10.0, "🥇"))
    formatter.project_evaluated(Score(0.0, "🚧"))
    stdout = capsys.readouterr().out

    expected = """\
    🥇 \x1b[1mModel: model1\x1b[0m (score: 10.0)
        \x1b[1;33mWARN\x1b[0m (critical) tests.conftest.rule_severity_critical: Error

    Project score: \x1b[1m0.0\x1b[0m 🚧

    Error: project score too low, fail_project_under = 5.0
    """
    assert stdout == dedent(expected)



"""Unit tests for the JSON formatter."""

from typing import Type

from dbt_score.formatters.json_formatter import JSONFormatter
from dbt_score.rule import Rule, RuleViolation
from dbt_score.scoring import Score


def test_json_formatter(
    capsys,
    default_config,
    manifest_loader,
    model1,
    source1,
    rule_severity_low,
    rule_severity_medium,
    rule_severity_critical,
):
    """Ensure the formatter has the correct output after evaluation."""
    formatter = JSONFormatter(manifest_loader=manifest_loader, config=default_config)
    results: dict[Type[Rule], RuleViolation | Exception | None] = {
        rule_severity_low: None,
        rule_severity_medium: Exception("Oh noes"),
        rule_severity_critical: RuleViolation("Error"),
    }
    formatter.evaluable_evaluated(model1, results, Score(10.0, "🥇"))
    formatter.evaluable_evaluated(source1, results, Score(10.0, "🥇"))
    formatter.project_evaluated(Score(10.0, "🥇"))
    stdout = capsys.readouterr().out
    assert (
        stdout
        == """{
  "evaluables": {
    "model1": {
      "score": 10.0,
      "badge": "🥇",
      "pass": true,
      "results": {
        "tests.conftest.rule_severity_low": {
          "result": "OK",
          "severity": "low",
          "message": null
        },
        "tests.conftest.rule_severity_medium": {
          "result": "ERR",
          "severity": "medium",
          "message": "Oh noes"
        },
        "tests.conftest.rule_severity_critical": {
          "result": "WARN",
          "severity": "critical",
          "message": "Error"
        }
      },
      "type": "model"
    },
    "table1": {
      "score": 10.0,
      "badge": "🥇",
      "pass": true,
      "results": {
        "tests.conftest.rule_severity_low": {
          "result": "OK",
          "severity": "low",
          "message": null
        },
        "tests.conftest.rule_severity_medium": {
          "result": "ERR",
          "severity": "medium",
          "message": "Oh noes"
        },
        "tests.conftest.rule_severity_critical": {
          "result": "WARN",
          "severity": "critical",
          "message": "Error"
        }
      },
      "type": "source"
    }
  },
  "project": {
    "score": 10.0,
    "badge": "🥇",
    "pass": true
  }
}
"""
    )



"""Test configuration."""

import json
from pathlib import Path
from typing import Any, Type

from dbt_score import Model, Rule, RuleViolation, Seed, Severity, Snapshot, Source, rule
from dbt_score.config import Config
from dbt_score.models import ManifestLoader
from dbt_score.rule_filter import RuleFilter, rule_filter
from pytest import fixture

# Configuration


@fixture()
def default_config() -> Config:
    """Return a default Config object."""
    return Config()


@fixture
def valid_config_path() -> Path:
    """Return the path of the configuration."""
    return Path(__file__).parent / "resources" / "pyproject.toml"


@fixture
def invalid_config_path() -> Path:
    """Return the path of the configuration."""
    return Path(__file__).parent / "resources" / "invalid_pyproject.toml"


# Manifest


@fixture
def manifest_empty_path() -> Path:
    """Return the path of an empty manifest."""
    return Path(__file__).parent / "resources" / "manifest_empty.json"


@fixture
def manifest_path() -> Path:
    """Return the path of a manifest."""
    return Path(__file__).parent / "resources" / "manifest.json"


@fixture
def raw_manifest(manifest_path) -> Any:
    """Return a raw manifest."""
    return json.loads(manifest_path.read_text(encoding="utf-8"))


@fixture
def manifest_loader(manifest_path) -> ManifestLoader:
    """Return an instantiated and loaded manifest loader."""
    return ManifestLoader(file_path=manifest_path)


# Models


@fixture
def model1(raw_manifest) -> Model:
    """Model 1."""
    return Model.from_node(raw_manifest["nodes"]["model.package.model1"], [])


@fixture
def model2(raw_manifest) -> Model:
    """Model 2."""
    return Model.from_node(raw_manifest["nodes"]["model.package.model2"], [])


# Sources


@fixture
def source1(raw_manifest) -> Source:
    """Source 1."""
    return Source.from_node(
        raw_manifest["sources"]["source.package.my_source.table1"], []
    )


@fixture
def source2(raw_manifest) -> Source:
    """Source 2."""
    return Source.from_node(
        raw_manifest["sources"]["source.package.my_source.table2"], []
    )


# Snapshots


@fixture
def snapshot1(raw_manifest) -> Snapshot:
    """Snapshot 1."""
    return Snapshot.from_node(raw_manifest["nodes"]["snapshot.package.snapshot1"], [])


@fixture
def snapshot2(raw_manifest) -> Snapshot:
    """Snapshot 2."""
    return Snapshot.from_node(raw_manifest["nodes"]["snapshot.package.snapshot2"], [])


# Multiple ways to create rules


@fixture
def decorator_rule() -> Type[Rule]:
    """An example rule created with the rule decorator."""

    @rule()
    def example_rule(model: Model) -> RuleViolation | None:
        """Description of the rule."""
        if model.name == "model1":
            return RuleViolation(message="Model1 is a violation.")

    return example_rule


@fixture
def decorator_rule_no_parens() -> Type[Rule]:
    """An example rule created with the rule decorator without parentheses."""

    @rule
    def example_rule(model: Model) -> RuleViolation | None:
        """Description of the rule."""
        if model.name == "model1":
            return RuleViolation(message="Model1 is a violation.")

    return example_rule


@fixture
def decorator_rule_args() -> Type[Rule]:
    """An example rule created with the rule decorator with arguments."""

    @rule(description="Description of the rule.")
    def example_rule(model: Model) -> RuleViolation | None:
        if model.name == "model1":
            return RuleViolation(message="Model1 is a violation.")

    return example_rule


@fixture
def class_rule() -> Type[Rule]:
    """An example rule created with a class."""

    class ExampleRule(Rule):
        """Example rule."""

        description = "Description of the rule."

        def evaluate(self, model: Model) -> RuleViolation | None:  # type: ignore[override]
            """Evaluate model."""
            if model.name == "model1":
                return RuleViolation(message="Model1 is a violation.")

    return ExampleRule


@fixture
def decorator_rule_snapshot() -> Type[Rule]:
    """An example rule created with the rule decorator."""

    @rule()
    def example_rule_snapshot(snapshot: Snapshot) -> RuleViolation | None:
        """Description of the rule."""
        if snapshot.name == "snapshot`":
            return RuleViolation(message="Snapshot1 is a violation.")

    return example_rule_snapshot


@fixture
def decorator_rule_no_parens_snapshot() -> Type[Rule]:
    """An example rule created with the rule decorator without parentheses."""

    @rule
    def example_rule(snapshot: Snapshot) -> RuleViolation | None:
        """Description of the rule."""
        if snapshot.name == "snapshot1":
            return RuleViolation(message="Snapshot1 is a violation.")

    return example_rule


@fixture
def decorator_rule_args_snapshot() -> Type[Rule]:
    """An example rule created with the rule decorator with arguments."""

    @rule(description="Description of the rule.")
    def example_rule(snapshot: Snapshot) -> RuleViolation | None:
        if snapshot.name == "Snapshot1":
            return RuleViolation(message="Snapshot1 is a violation.")

    return example_rule


@fixture
def class_rule_snapshot() -> Type[Rule]:
    """An example rule created with a class."""

    class ExampleRule(Rule):
        """Example rule."""

        description = "Description of the rule."

        def evaluate(self, snapshot: Snapshot) -> RuleViolation | None:  # type: ignore[override]
            """Evaluate snapshot."""
            if snapshot.name == "snapshot1":
                return RuleViolation(message="Snapshot1 is a violation.")

    return ExampleRule


@fixture
def decorator_rule_source() -> Type[Rule]:
    """An example rule created with the rule decorator."""

    @rule()
    def example_rule_source(source: Source) -> RuleViolation | None:
        """Description of the rule."""
        if source.name == "table1":
            return RuleViolation(message="Source1 is a violation.")

    return example_rule_source


@fixture
def decorator_rule_no_parens_source() -> Type[Rule]:
    """An example rule created with the rule decorator without parentheses."""

    @rule
    def example_rule_source(source: Source) -> RuleViolation | None:
        """Description of the rule."""
        if source.name == "table1":
            return RuleViolation(message="Source1 is a violation.")

    return example_rule_source


@fixture
def decorator_rule_args_source() -> Type[Rule]:
    """An example rule created with the rule decorator with arguments."""

    @rule(description="Description of the rule.")
    def example_rule_source(source: Source) -> RuleViolation | None:
        if source.name == "table1":
            return RuleViolation(message="Source1 is a violation.")

    return example_rule_source


@fixture
def class_rule_source() -> Type[Rule]:
    """An example rule created with a class."""

    class ExampleRuleSource(Rule):
        """Example rule."""

        description = "Description of the rule."

        def evaluate(self, source: Source) -> RuleViolation | None:  # type: ignore[override]
            """Evaluate source."""
            if source.name == "table1":
                return RuleViolation(message="Source1 is a violation.")

    return ExampleRuleSource


# Rules


@fixture
def rule_severity_low() -> Type[Rule]:
    """An example rule with LOW severity."""

    @rule(severity=Severity.LOW)
    def rule_severity_low(model: Model) -> RuleViolation | None:
        """Rule with LOW severity."""
        if model.name != "model1":
            return RuleViolation(message="Linting error")

    return rule_severity_low


@fixture
def rule_severity_medium() -> Type[Rule]:
    """An example rule with MEDIUM severity."""

    @rule(severity=Severity.MEDIUM)
    def rule_severity_medium(model: Model) -> RuleViolation | None:
        """Rule with MEDIUM severity."""
        if model.name != "model1":
            return RuleViolation(message="Linting error")

    return rule_severity_medium


@fixture
def rule_severity_high() -> Type[Rule]:
    """An example rule with HIGH severity."""

    @rule(severity=Severity.HIGH)
    def rule_severity_high(model: Model) -> RuleViolation | None:
        """Rule with HIGH severity."""
        if model.name != "model1":
            return RuleViolation(message="Linting error")

    return rule_severity_high


@fixture
def rule_severity_critical() -> Type[Rule]:
    """An example rule with CRITICAL severity."""

    @rule(severity=Severity.CRITICAL)
    def rule_severity_critical(model: Model) -> RuleViolation | None:
        """Rule with CRITICAL severity."""
        if model.name != "model1":
            return RuleViolation(message="Linting error")

    return rule_severity_critical


@fixture
def rule_with_config() -> Type[Rule]:
    """An example rule with additional configuration."""

    @rule
    def rule_with_config(
        model: Model, model_name: str = "model1"
    ) -> RuleViolation | None:
        """Rule with additional configuration."""
        if model.name != model_name:
            return RuleViolation(message=model_name)

    return rule_with_config


@fixture
def rule_error() -> Type[Rule]:
    """An example rule which fails to run."""

    @rule
    def rule_error(model: Model) -> RuleViolation | None:
        """Always failing rule."""
        raise Exception("Oh noes, something went wrong")

    return rule_error


@fixture
def model_rule_with_filter() -> Type[Rule]:
    """An example rule that skips through a filter."""

    @rule_filter
    def skip_model1(model: Model) -> bool:
        """Skips for model1, passes for model2."""
        return model.name != "model1"

    @rule(rule_filters={skip_model1()})
    def model_rule_with_filter(model: Model) -> RuleViolation | None:
        """Rule that always fails when not filtered."""
        return RuleViolation(message="I always fail.")

    return model_rule_with_filter


@fixture
def source_rule_with_filter() -> Type[Rule]:
    """An example rule that skips through a filter."""

    @rule_filter
    def skip_source1(source: Source) -> bool:
        """Skips for source1, passes for source2."""
        return source.name != "table1"

    @rule(rule_filters={skip_source1()})
    def source_rule_with_filter(source: Source) -> RuleViolation | None:
        """Rule that always fails when not filtered."""
        return RuleViolation(message="I always fail.")

    return source_rule_with_filter


@fixture
def snapshot_rule_with_filter() -> Type[Rule]:
    """An example rule that skips through a filter."""

    @rule_filter
    def skip_snapshot1(snapshot: Snapshot) -> bool:
        """Skips for snapshot1, passes for snapshot2."""
        return snapshot.name != "snapshot1"

    @rule(rule_filters={skip_snapshot1()})
    def snapshot_rule_with_filter(snapshot: Snapshot) -> RuleViolation | None:
        """Rule that always fails when not filtered."""
        return RuleViolation(message="I always fail.")

    return snapshot_rule_with_filter


@fixture
def model_class_rule_with_filter() -> Type[Rule]:
    """Using class definitions for filters and rules."""

    class SkipModel1(RuleFilter):
        description = "Filter defined by a class."

        def evaluate(self, model: Model) -> bool:  # type: ignore[override]
            """Skips for model1, passes for model2."""
            return model.name != "model1"

    class ModelRuleWithFilter(Rule):
        description = "Filter defined by a class."
        rule_filters = frozenset({SkipModel1()})

        def evaluate(self, model: Model) -> RuleViolation | None:  # type: ignore[override]
            return RuleViolation(message="I always fail.")

    return ModelRuleWithFilter


@fixture
def source_class_rule_with_filter() -> Type[Rule]:
    """Using class definitions for filters and rules."""

    class SkipSource1(RuleFilter):
        description = "Filter defined by a class."

        def evaluate(self, source: Source) -> bool:  # type: ignore[override]
            """Skips for source1, passes for source2."""
            return source.name != "table1"

    class SourceRuleWithFilter(Rule):
        description = "Filter defined by a class."
        rule_filters = frozenset({SkipSource1()})

        def evaluate(self, source: Source) -> RuleViolation | None:  # type: ignore[override]
            return RuleViolation(message="I always fail.")

    return SourceRuleWithFilter


@fixture
def snapshot_class_rule_with_filter() -> Type[Rule]:
    """Using class definitions for filters and rules."""

    class SkipSnapshot1(RuleFilter):
        description = "Filter defined by a class."

        def evaluate(self, snapshot: Snapshot) -> bool:  # type: ignore[override]
            """Skips for snapshot1, passes for snapshot2."""
            return snapshot.name != "snapshot1"

    class SnapshotRuleWithFilter(Rule):
        description = "Filter defined by a class."
        rule_filters = frozenset({SkipSnapshot1()})

        def evaluate(self, snapshot: Snapshot) -> RuleViolation | None:  # type: ignore[override]
            return RuleViolation(message="I always fail.")

    return SnapshotRuleWithFilter


@fixture
def seed1(raw_manifest) -> Seed:
    """Seed 1."""
    return Seed.from_node(raw_manifest["nodes"]["seed.package.seed1"], [])


@fixture
def seed2(raw_manifest) -> Seed:
    """Seed 2."""
    return Seed.from_node(raw_manifest["nodes"]["seed.package.seed2"], [])


@fixture
def decorator_rule_seed() -> Type[Rule]:
    """An example rule created with the rule decorator."""

    @rule()
    def example_rule_seed(seed: Seed) -> RuleViolation | None:
        """Description of the rule."""
        if seed.name == "seed1":
            return RuleViolation(message="Seed1 is a violation.")

    return example_rule_seed


@fixture
def decorator_rule_no_parens_seed() -> Type[Rule]:
    """An example rule created with the rule decorator without parentheses."""

    @rule
    def example_rule(seed: Seed) -> RuleViolation | None:
        """Description of the rule."""
        if seed.name == "seed1":
            return RuleViolation(message="Seed1 is a violation.")

    return example_rule


@fixture
def decorator_rule_args_seed() -> Type[Rule]:
    """An example rule created with the rule decorator with arguments."""

    @rule(description="Description of the rule.")
    def example_rule(seed: Seed) -> RuleViolation | None:
        if seed.name == "seed1":
            return RuleViolation(message="Seed1 is a violation.")

    return example_rule


@fixture
def class_rule_seed() -> Type[Rule]:
    """An example rule created with a class."""

    class ExampleRule(Rule):
        """Example rule."""

        description = "Description of the rule."

        def evaluate(self, seed: Seed) -> RuleViolation | None:  # type: ignore[override]
            """Evaluate seed."""
            if seed.name == "seed1":
                return RuleViolation(message="Seed1 is a violation.")

    return ExampleRule


@fixture
def seed_rule_with_filter() -> Type[Rule]:
    """An example rule that skips through a filter."""

    @rule_filter
    def skip_seed1(seed: Seed) -> bool:
        """Skips for seed1, passes for seed2."""
        return seed.name != "seed1"

    @rule(rule_filters={skip_seed1()})
    def seed_rule_with_filter(seed: Seed) -> RuleViolation | None:
        """Rule that always fails when not filtered."""
        return RuleViolation(message="I always fail.")

    return seed_rule_with_filter


@fixture
def seed_class_rule_with_filter() -> Type[Rule]:
    """Using class definitions for filters and rules."""

    class SkipSeed1(RuleFilter):
        description = "Filter defined by a class."

        def evaluate(self, seed: Seed) -> bool:  # type: ignore[override]
            """Skips for seed1, passes for seed2."""
            return seed.name != "seed1"

    class SeedRuleWithFilter(Rule):
        description = "Filter defined by a class."
        rule_filters = frozenset({SkipSeed1()})

        def evaluate(self, seed: Seed) -> RuleViolation | None:  # type: ignore[override]
            return RuleViolation(message="I always fail.")

    return SeedRuleWithFilter



"""Tests for seed rules."""

from dbt_score.rule import RuleViolation
from dbt_score.rules.generic import (
    seed_columns_have_description,
    seed_has_description,
    seed_has_owner,
)


def test_seed_has_description(seed1, seed2):
    """Test seed_has_description rule."""
    # seed1 has a description, seed2 doesn't
    rule = seed_has_description()
    assert rule.evaluate(seed1) is None
    assert isinstance(rule.evaluate(seed2), RuleViolation)


def test_seed_columns_have_description(seed1, seed2):
    """Test seed_columns_have_description rule."""
    # seed1's column has a description, seed2's column doesn't
    rule = seed_columns_have_description()
    assert rule.evaluate(seed1) is None
    assert isinstance(rule.evaluate(seed2), RuleViolation)


def test_seed_has_owner(seed1, seed2):
    """Test seed_has_owner rule."""
    # Add owner to seed1, not to seed2
    seed1.config["meta"] = {"owner": "Data Team"}

    rule = seed_has_owner()
    assert rule.evaluate(seed1) is None
    assert isinstance(rule.evaluate(seed2), RuleViolation)



"""Unit tests for the manifest formatter."""

import json

from dbt_score.evaluation import EvaluableResultsType
from dbt_score.formatters.manifest_formatter import ManifestFormatter
from dbt_score.rule import RuleViolation
from dbt_score.scoring import Score


def test_manifest_formatter_model(
    capsys,
    default_config,
    manifest_loader,
    model1,
    rule_severity_low,
    rule_severity_medium,
    rule_severity_critical,
):
    """Ensure the formatter has the correct output after model evaluation."""
    formatter = ManifestFormatter(
        manifest_loader=manifest_loader, config=default_config
    )
    results: EvaluableResultsType = {
        rule_severity_low: None,
        rule_severity_medium: Exception("Oh noes"),
        rule_severity_critical: RuleViolation("Error"),
    }
    formatter.evaluable_evaluated(model1, results, Score(10.0, "🥇"))
    stdout = capsys.readouterr().out
    assert stdout == ""


def test_manifest_formatter_project(
    capsys,
    default_config,
    manifest_loader,
    model1,
    model2,
    source1,
    source2,
    rule_severity_low,
    rule_severity_medium,
    rule_severity_critical,
):
    """Ensure the formatter has the correct output after project evaluation."""
    formatter = ManifestFormatter(
        manifest_loader=manifest_loader, config=default_config
    )
    result1: EvaluableResultsType = {
        rule_severity_low: None,
        rule_severity_medium: Exception("Oh noes"),
        rule_severity_critical: RuleViolation("Error"),
    }
    result2: EvaluableResultsType = {
        rule_severity_low: None,
        rule_severity_medium: None,
        rule_severity_critical: None,
    }

    formatter.evaluable_evaluated(model1, result1, Score(5.0, "🚧"))
    formatter.evaluable_evaluated(model2, result2, Score(10.0, "🥇"))
    formatter.evaluable_evaluated(source1, result1, Score(5.0, "🚧"))
    formatter.evaluable_evaluated(source2, result2, Score(10.0, "🥇"))
    formatter.project_evaluated(Score(7.5, "🥉"))

    stdout = capsys.readouterr().out
    new_manifest = json.loads(stdout)
    assert new_manifest["nodes"]["model.package.model1"]["meta"]["score"] == 5.0
    assert new_manifest["nodes"]["model.package.model1"]["meta"]["badge"] == "🚧"
    assert new_manifest["nodes"]["model.package.model2"]["meta"]["score"] == 10.0
    assert new_manifest["nodes"]["model.package.model2"]["meta"]["badge"] == "🥇"

    assert (
        new_manifest["sources"]["source.package.my_source.table1"]["meta"]["score"]
        == 5.0
    )
    assert (
        new_manifest["sources"]["source.package.my_source.table1"]["meta"]["badge"]
        == "🚧"
    )
    assert (
        new_manifest["sources"]["source.package.my_source.table2"]["meta"]["score"]
        == 10.0
    )
    assert (
        new_manifest["sources"]["source.package.my_source.table2"]["meta"]["badge"]
        == "🥇"
    )



"""Unit tests for the scoring module."""


from dbt_score.rule import RuleViolation
from dbt_score.scoring import Score, Scorer


def test_scorer_model_no_results(default_config):
    """Test scorer with a model without any result."""
    scorer = Scorer(config=default_config)
    assert scorer.score_evaluable({}).value == 10.0


def test_scorer_model_severity_low(default_config, rule_severity_low):
    """Test scorer with a model and one low severity rule."""
    scorer = Scorer(config=default_config)
    assert scorer.score_evaluable({rule_severity_low: None}).value == 10.0
    assert scorer.score_evaluable({rule_severity_low: Exception()}).value == 10.0
    assert (
        round(
            scorer.score_evaluable({rule_severity_low: RuleViolation("error")}).value, 2
        )
        == 6.67
    )


def test_scorer_model_severity_medium(default_config, rule_severity_medium):
    """Test scorer with a model and one medium severity rule."""
    scorer = Scorer(config=default_config)
    assert scorer.score_evaluable({rule_severity_medium: None}).value == 10.0
    assert scorer.score_evaluable({rule_severity_medium: Exception()}).value == 10.0
    assert (
        round(
            scorer.score_evaluable(
                {rule_severity_medium: RuleViolation("error")}
            ).value,
            2,
        )
        == 3.33
    )


def test_scorer_model_severity_high(default_config, rule_severity_high):
    """Test scorer with a model and one high severity rule."""
    scorer = Scorer(config=default_config)
    assert scorer.score_evaluable({rule_severity_high: None}).value == 10.0
    assert scorer.score_evaluable({rule_severity_high: Exception()}).value == 10.0
    assert (
        scorer.score_evaluable({rule_severity_high: RuleViolation("error")}).value
        == 0.0
    )


def test_scorer_model_severity_critical(default_config, rule_severity_critical):
    """Test scorer with a model and one critical severity rule."""
    scorer = Scorer(config=default_config)
    assert scorer.score_evaluable({rule_severity_critical: None}).value == 10.0
    assert scorer.score_evaluable({rule_severity_critical: Exception()}).value == 10.0
    assert (
        scorer.score_evaluable({rule_severity_critical: RuleViolation("error")}).value
        == 0.0
    )


def test_scorer_model_severity_critical_overwrites(
    default_config, rule_severity_low, rule_severity_critical
):
    """Test scorer with a model and multiple rules including one critical."""
    scorer = Scorer(config=default_config)
    assert (
        scorer.score_evaluable(
            {rule_severity_low: None, rule_severity_critical: RuleViolation("error")}
        ).value
        == 0.0
    )


def test_scorer_model_multiple_rules(
    default_config, rule_severity_low, rule_severity_medium, rule_severity_high
):
    """Test scorer with a model and multiple rules."""
    scorer = Scorer(config=default_config)
    assert (
        round(
            scorer.score_evaluable(
                {
                    rule_severity_low: None,
                    rule_severity_medium: Exception(),
                    rule_severity_high: RuleViolation("error"),
                }
            ).value,
            2,
        )
        == 6.67
    )

    assert (
        round(
            scorer.score_evaluable(
                {
                    rule_severity_low: Exception(),
                    rule_severity_medium: RuleViolation("error"),
                    rule_severity_high: None,
                }
            ).value,
            2,
        )
        == 7.78
    )

    assert (
        round(
            scorer.score_evaluable(
                {
                    rule_severity_low: RuleViolation("error"),
                    rule_severity_medium: Exception(),
                    rule_severity_high: None,
                }
            ).value,
            2,
        )
        == 8.89
    )


def test_scorer_aggregate_empty(default_config):
    """Test scorer aggregation with no results."""
    scorer = Scorer(config=default_config)
    assert scorer.score_aggregate_evaluables([]).value == 10.0


def test_scorer_aggregate_with_0(default_config):
    """Test scorer aggregation with one result that is 0.0."""
    scorer = Scorer(config=default_config)
    scores = [Score(1.0, ""), Score(5.0, ""), Score(0.0, "")]
    assert scorer.score_aggregate_evaluables(scores).value == 0.0


def test_scorer_aggregate_single(default_config):
    """Test scorer aggregation with a single results."""
    scorer = Scorer(config=default_config)
    assert scorer.score_aggregate_evaluables([Score(4.2, "")]).value == 4.2


def test_scorer_aggregate_multiple(default_config):
    """Test scorer aggregation with multiple results."""
    scorer = Scorer(config=default_config)
    assert (
        scorer.score_aggregate_evaluables(
            [Score(1.0, ""), Score(1.0, ""), Score(1.0, "")]
        ).value
        == 1.0
    )
    assert (
        scorer.score_aggregate_evaluables(
            [Score(1.0, ""), Score(7.4, ""), Score(4.2, "")]
        ).value
        == 4.2
    )
    assert (
        scorer.score_aggregate_evaluables(
            [Score(0.0, ""), Score(0.0, ""), Score(0.0, "")]
        ).value
        == 0.0
    )


def test_scorer_badge(default_config):
    """Test scorer awarding a badge."""
    scorer = Scorer(config=default_config)
    assert scorer._badge(10.0) == scorer._config.badge_config.first.icon
    assert scorer._badge(8.0) == scorer._config.badge_config.second.icon
    assert scorer._badge(7.0) == scorer._config.badge_config.third.icon
    assert scorer._badge(1.0) == scorer._config.badge_config.wip.icon



"""This module is responsible for loading configuration."""

import logging
import sys
from dataclasses import dataclass, field, replace
from pathlib import Path
from typing import Any, Final

from dbt_score.rule import RuleConfig

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib

logger = logging.getLogger(__name__)

DEFAULT_CONFIG_FILE = "pyproject.toml"


@dataclass
class Badge:
    """Badge object."""

    icon: str
    threshold: float


@dataclass
class BadgeConfig:
    """Configuration for badges."""

    third: Badge = field(default_factory=lambda: Badge("🥉", 6.0))
    second: Badge = field(default_factory=lambda: Badge("🥈", 8.0))
    first: Badge = field(default_factory=lambda: Badge("🥇", 10.0))
    wip: Badge = field(default_factory=lambda: Badge("🚧", 0.0))

    def validate(self) -> None:
        """Validate the badge configuration."""
        if not (self.first.threshold > self.second.threshold > self.third.threshold):
            raise ValueError("Invalid badge thresholds.")
        if self.first.threshold > 10.0:  # noqa: PLR2004 [magic-value-comparison]
            raise ValueError("first threshold must 10.0 or lower.")
        if self.third.threshold < 0.0:
            raise ValueError("third threshold must be 0.0 or higher.")
        if self.wip.threshold != 0.0:
            raise AttributeError("wip badge cannot have a threshold configuration.")


class Config:
    """Configuration for dbt-score."""

    _main_section: Final[str] = "tool.dbt-score"
    _options: Final[list[str]] = [
        "rule_namespaces",
        "disabled_rules",
        "inject_cwd_in_python_path",
        "fail_project_under",
        "fail_any_item_under",
        "show",
        "debug",
    ]
    _rules_section: Final[str] = "rules"
    _badges_section: Final[str] = "badges"

    def __init__(self) -> None:
        """Initialize the Config object."""
        self.rule_namespaces: list[str] = ["dbt_score.rules", "dbt_score_rules"]
        self.disabled_rules: list[str] = []
        self.inject_cwd_in_python_path = True
        self.rules_config: dict[str, RuleConfig] = {}
        self.config_file: Path | None = None
        self.badge_config: BadgeConfig = BadgeConfig()
        self.fail_project_under: float = 5.0
        self.fail_any_item_under: float = 5.0
        self.show: str = "failing-rules"
        self.debug: bool = False

    def set_option(self, option: str, value: Any) -> None:
        """Set an option in the config."""
        setattr(self, option, value)

    def _load_toml_file(self, file: str) -> None:
        """Load the options from a TOML file."""
        with open(file, "rb") as f:
            toml_data = tomllib.load(f)

        tools = toml_data.get("tool", {})
        dbt_score_config = tools.get("dbt-score", {})
        rules_config = dbt_score_config.pop(self._rules_section, {})
        badge_config = dbt_score_config.pop(self._badges_section, {})

        # Main configuration
        for option, value in dbt_score_config.items():
            if option in self._options:
                self.set_option(option, value)
            elif not isinstance(
                value, dict
            ):  # If value is a dictionary, it's another section
                logger.warning(
                    f"Option {option} in {self._main_section} not supported."
                )

        # Badge configuration
        for name, config in badge_config.items():
            try:
                default_config = getattr(self.badge_config, name)
                updated_config = replace(default_config, **config)
                setattr(self.badge_config, name, updated_config)
            except AttributeError as e:
                options = list(BadgeConfig.__annotations__.keys())
                raise AttributeError(f"Config only accepts badges: {options}.") from e
            except TypeError as e:
                options = list(Badge.__annotations__.keys())
                if name == "wip":
                    options.remove("threshold")
                raise AttributeError(
                    f"Badge {name}: config only accepts {options}."
                ) from e

        self.badge_config.validate()

        # Rule configuration
        self.rules_config = {
            name: RuleConfig.from_dict(config) for name, config in rules_config.items()
        }

    @staticmethod
    def get_config_file(directory: Path) -> Path | None:
        """Get the config file."""
        candidates = [directory]
        candidates.extend(directory.parents)
        for path in candidates:
            config_file = path / DEFAULT_CONFIG_FILE
            if config_file.exists():
                return config_file

    def load(self) -> None:
        """Load the config."""
        config_file = self.get_config_file(Path.cwd())
        if config_file:
            self._load_toml_file(str(config_file))

    def overload(self, values: dict[str, Any]) -> None:
        """Overload config with additional values."""
        for key, value in values.items():
            self.set_option(key, value)



"""Tests for the module config_parser."""
from pathlib import Path
from unittest.mock import patch

import pytest
from dbt_score.config import BadgeConfig, Config
from dbt_score.rule import RuleConfig, Severity


def test_load_valid_toml_file(valid_config_path):
    """Test that a valid config file loads correctly."""
    config = Config()
    config._load_toml_file(str(valid_config_path))
    assert config.rule_namespaces == ["foo", "tests"]
    assert config.disabled_rules == ["foo.foo", "tests.bar"]
    assert config.rules_config["foo.bar"].severity == Severity.CRITICAL
    assert (
        config.rules_config["tests.rules.rules.rule_test_example"].severity
        == Severity.CRITICAL
    )
    assert config.badge_config.third.threshold == 6.5
    assert config.badge_config.second.threshold == 7.5
    assert config.badge_config.first.threshold == 9.5
    assert config.badge_config.wip.icon == "🏗️"
    assert config.badge_config.third.icon == "3️⃣"
    assert config.badge_config.second.icon == "2️⃣"
    assert config.badge_config.first.icon == "1️⃣"
    assert config.fail_project_under == 7.5
    assert config.fail_any_item_under == 6.9
    assert config.rules_config[
        "tests.rules.rules.rule_test_example"
    ].rule_filter_names == ["tests.rules.rule_filters.skip_model1"]


def test_load_invalid_toml_file(caplog, invalid_config_path):
    """Test that an invalid config file logs a warning."""
    config = Config()
    config._load_toml_file(str(invalid_config_path))
    assert "Option foo in tool.dbt-score not supported." in caplog.text


def test_invalid_rule_config(rule_severity_low):
    """Test that an invalid rule config raises an exception."""
    config = RuleConfig(config={"foo": "bar"})
    with pytest.raises(
        AttributeError,
        match="Unknown rule parameter: foo for rule "
        "tests.conftest.rule_severity_low.",
    ):
        rule_severity_low(config)


@patch("dbt_score.config.open")
def test_load_invalid_badge_config(mock_open):
    """Test that an invalid badge config raises an exception."""
    config = Config()

    with patch("dbt_score.config.tomllib.load") as mock_load:
        mock_load.return_value = {
            "tool": {"dbt-score": {"badges": {"foo": {"threshold": 1.0}}}}
        }
        with pytest.raises(AttributeError, match="Config only accepts badges:"):
            config._load_toml_file("foo")

        mock_load.return_value = {
            "tool": {"dbt-score": {"badges": {"first": {"foo": "bar"}}}}
        }
        with pytest.raises(AttributeError, match="Badge first: config only accepts"):
            config._load_toml_file("foo")


def test_invalid_badge_thresholds():
    """Test that invalid badge thresholds raises an exception."""
    badge_config = BadgeConfig()
    badge_config.third.threshold = 9.0
    badge_config.second.threshold = 8.0
    badge_config.first.threshold = 10.0
    with pytest.raises(ValueError, match="Invalid badge thresholds."):
        badge_config.validate()

    badge_config = BadgeConfig()
    badge_config.third.threshold = 8.0
    badge_config.second.threshold = 9.5
    badge_config.first.threshold = 9.5
    with pytest.raises(ValueError, match="Invalid badge thresholds."):
        badge_config.validate()

    badge_config = BadgeConfig()
    badge_config.third.threshold = -1
    with pytest.raises(ValueError, match="third threshold must be 0.0 or higher."):
        badge_config.validate()

    badge_config = BadgeConfig()
    badge_config.first.threshold = 11.0
    with pytest.raises(ValueError, match="first threshold must 10.0 or lower."):
        badge_config.validate()

    badge_config = BadgeConfig()
    badge_config.wip.threshold = 1.0
    with pytest.raises(
        AttributeError, match="wip badge cannot have a threshold configuration."
    ):
        badge_config.validate()


def test_valid_rule_config(valid_config_path, rule_with_config):
    """Test that a valid rule config can be loaded."""
    config = RuleConfig(severity=Severity(4), config={"model_name": "baz"})
    rule_with_config = rule_with_config(config)
    assert rule_with_config.severity == Severity.CRITICAL
    assert rule_with_config.default_config == {"model_name": "model1"}
    assert rule_with_config.config == {"model_name": "baz"}


def test_get_config_file():
    """Test that the config file is found in the current directory."""
    directory = Path(__file__).parent / "resources"
    config = Config()
    config_file = config.get_config_file(directory)
    assert config_file == directory / "pyproject.toml"


def test_get_parent_config_file():
    """Test that the config file is found in the parent directory."""
    directory = Path(__file__).parent / "resources" / "sub_dir"
    config = Config()
    config_file = config.get_config_file(directory)
    assert config_file == directory.parent / "pyproject.toml"


def test_config_overload(valid_config_path):
    """Test overloading of config values."""
    config = Config()
    config._load_toml_file(str(valid_config_path))
    config.overload({"rule_namespaces": ["x", "y"], "disabled_rules": ["foo"]})
    assert config.rule_namespaces == ["x", "y"]
    assert config.disabled_rules == ["foo"]



"""Entry point of the dbt_score.

This enables module to be run directly.
"""

import logging
import sys

from dbt_score.cli import cli


def set_logging() -> None:
    """Set logging configuration."""
    log_format = "%(asctime)s %(levelname)s [%(name)s] %(message)s"
    handler: logging.Handler = logging.StreamHandler(sys.stderr)
    logging.basicConfig(format=log_format, handlers=[handler])
    for handler in logging.getLogger().handlers:
        handler.setLevel(logging.WARNING)


def main() -> None:
    """Main entrypoint."""
    set_logging()
    cli()


if __name__ == "__main__":
    main()



"""This module is responsible for evaluating rules."""

from __future__ import annotations

import pdb
import traceback
from itertools import chain
from typing import Type, cast

from dbt_score.config import Config
from dbt_score.formatters import Formatter
from dbt_score.models import Evaluable, ManifestLoader
from dbt_score.rule import Rule, RuleViolation
from dbt_score.rule_registry import RuleRegistry
from dbt_score.scoring import Score, Scorer

# The results of a given evaluable are stored in a dictionary, mapping rules to either:
# - None if there was no issue
# - A RuleViolation if a linting error was found
# - An Exception if the rule failed to run
EvaluableResultsType = dict[Type[Rule], None | RuleViolation | Exception]


class Evaluation:
    """Evaluate a set of rules on a set of nodes."""

    def __init__(
        self,
        rule_registry: RuleRegistry,
        manifest_loader: ManifestLoader,
        formatter: Formatter,
        scorer: Scorer,
        config: Config,
    ) -> None:
        """Create an Evaluation object.

        Args:
            rule_registry: A rule registry to access rules.
            manifest_loader: A manifest loader to access dbt metadata.
            formatter: A formatter to display results.
            scorer: A scorer to compute scores.
            config: A configuration.
        """
        self._rule_registry = rule_registry
        self._manifest_loader = manifest_loader
        self._formatter = formatter
        self._scorer = scorer
        self._config = config

        # For each evaluable, its results
        self.results: dict[Evaluable, EvaluableResultsType] = {}

        # For each evaluable, its computed score
        self.scores: dict[Evaluable, Score] = {}

        # The aggregated project score
        self.project_score: Score

    def evaluate(self) -> None:
        """Evaluate all rules."""
        rules = self._rule_registry.rules.values()

        for evaluable in chain(
            self._manifest_loader.models.values(),
            self._manifest_loader.sources.values(),
            self._manifest_loader.snapshots.values(),
            self._manifest_loader.seeds.values(),
        ):
            # type inference on elements from `chain` is wonky
            # and resolves to superclass HasColumnsMixin
            evaluable = cast(Evaluable, evaluable)
            self.results[evaluable] = {}
            for rule in rules:
                try:
                    if rule.should_evaluate(evaluable):
                        result = rule.evaluate(evaluable, **rule.config)
                        self.results[evaluable][rule.__class__] = result
                except Exception as e:
                    if self._config.debug:
                        traceback.print_exc()
                        pdb.post_mortem()
                    self.results[evaluable][rule.__class__] = e

            self.scores[evaluable] = self._scorer.score_evaluable(
                self.results[evaluable]
            )
            self._formatter.evaluable_evaluated(
                evaluable, self.results[evaluable], self.scores[evaluable]
            )

        # Compute score for project
        self.project_score = self._scorer.score_aggregate_evaluables(
            list(self.scores.values())
        )

        # Add null check before calling project_evaluated
        if (
            self._manifest_loader.models
            or self._manifest_loader.sources
            or self._manifest_loader.snapshots
            or self._manifest_loader.seeds
        ):
            self._formatter.project_evaluated(self.project_score)



"""CLI interface."""

import logging
import traceback
from pathlib import Path
from typing import Final, Literal

import click
from click.core import ParameterSource

from dbt_score.config import Config
from dbt_score.dbt_utils import (
    DbtParseException,
    dbt_parse,
    get_default_manifest_path,
)
from dbt_score.lint import lint_dbt_project
from dbt_score.rule_catalog import display_catalog

logger = logging.getLogger(__name__)

BANNER: Final[str] = r"""
          __ __     __
     ____/ // /_   / /_        _____ _____ ____   _____ ___
    / __  // __ \ / __/______ / ___// ___// __ \ / ___// _ \
   / /_/ // /_/ // /_ /_____/(__  )/ /__ / /_/ // /   /  __/
   \__,_//_.___/ \__/       /____/ \___/ \____//_/    \___/
    """


@click.version_option(message="%(version)s")
@click.group(
    help=f"\b{BANNER}",
    invoke_without_command=False,
    context_settings={"help_option_names": ["-h", "--help"]},
)
def cli() -> None:
    """CLI entrypoint."""


@cli.command()
@click.option(
    "--format",
    "-f",
    help="Output format. Plain is suitable for terminals, manifest for rich "
    "documentation, json for machine-readable output.",
    type=click.Choice(["plain", "manifest", "ascii", "json"]),
    default="plain",
)
@click.option(
    "--select",
    "-s",
    help="Specify the nodes to include.",
    multiple=True,
)
@click.option(
    "--namespace",
    "-n",
    help="Namespace to look for rules.",
    default=None,
    multiple=True,
)
@click.option(
    "--disabled-rule",
    help="Rule to disable.",
    default=None,
    multiple=True,
)
@click.option(
    "--manifest",
    "-m",
    help="Manifest filepath.",
    type=click.Path(path_type=Path),
    default=get_default_manifest_path(),
)
@click.option(
    "--run-dbt-parse",
    "-p",
    help="Run dbt parse.",
    is_flag=True,
    default=False,
)
@click.option(
    "--fail-project-under",
    help="Fail if the project score is under this value.",
    type=float,
    is_flag=False,
    default=None,
)
@click.option(
    "--fail-any-item-under",
    help="Fail if any evaluable item is under this value.",
    type=float,
    is_flag=False,
    default=None,
)
@click.option(
    "--show",
    help="Type of output which should be shown "
    "when using `plain` as `--format`. "
    "`all` shows all items and all rules. "
    "`failing-items` shows failing rules of failing items. "
    "`failing-rules` shows failing rules of all items. "
    "Default is --failing-rules.",
    type=click.Choice(["all", "failing-items", "failing-rules"]),
    is_flag=False,
    default="failing-rules",
)
@click.option(
    "--debug",
    "-d",
    help="Jump in a debugger in case of rule failure to evaluate.",
    type=bool,
    is_flag=True,
    default=False,
)
@click.pass_context
def lint(  # noqa: PLR0912, PLR0913, C901
    ctx: click.Context,
    format: Literal["plain", "manifest", "ascii"],
    select: tuple[str],
    namespace: list[str],
    disabled_rule: list[str],
    manifest: Path,
    run_dbt_parse: bool,
    fail_project_under: float,
    fail_any_item_under: float,
    show: Literal["all", "failing-items", "failing-rules"],
    debug: bool,
) -> None:
    """Lint dbt metadata."""
    manifest_provided = (
        click.get_current_context().get_parameter_source("manifest")
        != ParameterSource.DEFAULT
    )
    if manifest_provided and run_dbt_parse:
        raise click.UsageError("--run-dbt-parse cannot be used with --manifest.")

    config = Config()
    config.load()
    if namespace:
        config.overload({"rule_namespaces": namespace})
    if disabled_rule:
        config.overload({"disabled_rules": disabled_rule})
    if fail_project_under:
        config.overload({"fail_project_under": fail_project_under})
    if fail_any_item_under:
        config.overload({"fail_any_item_under": fail_any_item_under})
    if show:
        config.overload({"show": show})
    if debug:
        config.overload({"debug": debug})

    try:
        if run_dbt_parse:
            dbt_parse()
        evaluation = lint_dbt_project(
            manifest_path=manifest, config=config, format=format, select=select
        )

    except FileNotFoundError:
        logger.error(
            "dbt's manifest.json could not be found. If you're in a dbt project, be "
            "sure to run 'dbt parse' first, or use the option '--run-dbt-parse'."
        )
        ctx.exit(2)

    except DbtParseException:
        logger.error("dbt failed to parse project. Run `dbt parse` to investigate.")
        ctx.exit(2)

    except Exception:
        logger.error(traceback.format_exc())
        ctx.exit(2)

    if (
        any(x.value < config.fail_any_item_under for x in evaluation.scores.values())
        or evaluation.project_score.value < config.fail_project_under
    ):
        ctx.exit(1)


@cli.command(name="list")
@click.option(
    "--namespace",
    "-n",
    help="Namespace.",
    default=None,
    multiple=True,
)
@click.option(
    "--disabled-rule",
    help="Rule to disable.",
    default=None,
    multiple=True,
)
@click.option(
    "--title",
    help="Page title (Markdown only).",
    default=None,
)
@click.option(
    "--format",
    "-f",
    help="Output format.",
    type=click.Choice(["terminal", "markdown"]),
    default="terminal",
)
def list_command(
    namespace: list[str], disabled_rule: list[str], title: str, format: str
) -> None:
    """Display rules list."""
    config = Config()
    config.load()
    if namespace:
        config.overload({"rule_namespaces": namespace})
    if disabled_rule:
        config.overload({"disabled_rules": disabled_rule})

    display_catalog(config, title, format)



"""dbt-score exceptions."""


class DuplicatedRuleException(Exception):
    """Two rules with the same name are defined."""

    def __init__(self, rule_name: str):
        """Instantiate exception."""
        super().__init__(
            f"Rule {rule_name} is defined twice. Rules must have unique names."
        )



"""Init dbt_score package."""

from dbt_score.models import Model, Seed, Snapshot, Source
from dbt_score.rule import Rule, RuleViolation, Severity, rule
from dbt_score.rule_filter import RuleFilter, rule_filter

__all__ = [
    "Model",
    "Source",
    "Snapshot",
    "Seed",
    "RuleFilter",
    "Rule",
    "RuleViolation",
    "Severity",
    "rule_filter",
    "rule",
]



"""dbt utilities."""

import contextlib
import os
from functools import wraps
from pathlib import Path
from typing import Any, Callable, Iterable, Iterator, cast

# Conditionally import dbt objects.
try:
    DBT_INSTALLED = True
    from dbt.cli.main import dbtRunner, dbtRunnerResult  # type: ignore
except ImportError:
    DBT_INSTALLED = False


class DbtNotInstalledException(Exception):
    """Raised when trying to run dbt when dbt is not installed."""


class DbtParseException(Exception):
    """Raised when dbt parse fails."""


class DbtLsException(Exception):
    """Raised when dbt ls fails."""


def dbt_required(func: Callable[..., Any]) -> Callable[..., Any]:
    """Decorator for methods that require dbt to be installed."""

    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        if not DBT_INSTALLED:
            raise DbtNotInstalledException(
                "This option requires dbt to be installed in the same Python"
                "environment as dbt-score."
            )
        return func(*args, **kwargs)

    return wrapper


@contextlib.contextmanager
def _disable_dbt_stdout() -> Iterator[None]:
    with contextlib.redirect_stdout(None):
        yield


@dbt_required
def dbt_parse() -> "dbtRunnerResult":
    """Parse a dbt project.

    Returns:
        The dbt parse run result.

    Raises:
        DbtParseException: dbt parse failed.
    """
    with _disable_dbt_stdout():
        result: "dbtRunnerResult" = dbtRunner().invoke(["parse"])

    if not result.success:
        raise DbtParseException("dbt parse failed.") from result.exception

    return result


@dbt_required
def dbt_ls(select: Iterable[str] | None) -> Iterable[str]:
    """Run dbt ls."""
    cmd = [
        "ls",
        "--resource-types",
        "model",
        "source",
        "snapshot",
        "seed",
        "--output",
        "name",
    ]
    if select:
        cmd += ["--select", *select]

    with _disable_dbt_stdout():
        result: "dbtRunnerResult" = dbtRunner().invoke(cmd)

    if not result.success:
        raise DbtLsException("dbt ls failed.") from result.exception

    selected = cast(Iterable[str], result.result)  # mypy hint
    return selected


def get_default_manifest_path() -> Path:
    """Get the manifest path."""
    return (
        Path().cwd()
        / os.getenv("DBT_PROJECT_DIR", "")
        / os.getenv("DBT_TARGET_DIR", "target")
        / "manifest.json"
    )



"""Formatters are used to output CLI results."""

from __future__ import annotations

import typing
from abc import ABC, abstractmethod

from dbt_score.config import Config
from dbt_score.scoring import Score

if typing.TYPE_CHECKING:
    from dbt_score.evaluation import EvaluableResultsType
from dbt_score.models import Evaluable, ManifestLoader


class Formatter(ABC):
    """Abstract class to define a formatter."""

    def __init__(self, manifest_loader: ManifestLoader, config: Config):
        """Instantiate a formatter."""
        self._manifest_loader = manifest_loader
        self._config = config

    @abstractmethod
    def evaluable_evaluated(
        self, evaluable: Evaluable, results: EvaluableResultsType, score: Score
    ) -> None:
        """Callback when an evaluable item has been evaluated."""
        raise NotImplementedError

    @abstractmethod
    def project_evaluated(self, score: Score) -> None:
        """Callback when a project has been evaluated."""
        raise NotImplementedError



"""Rule filters."""

from dbt_score import rule_filter
from dbt_score.models import Model


@rule_filter
def is_table(model: Model) -> bool:
    """Models that are tables."""
    return model.config.get("materialized") in {"table", "incremental"}



"""Objects related to loading the dbt manifest."""

import json
import logging
import re
from collections import defaultdict
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Iterable, Literal, TypeAlias, Union

from dbt_score.dbt_utils import dbt_ls

logger = logging.getLogger(__name__)


@dataclass
class Constraint:
    """Constraint for a model or a column.

    Attributes:
        type: The type of the constraint, e.g. `foreign_key`.
        name: The name of the constraint.
        expression: The expression of the constraint, e.g. `schema.other_table`.
        columns: The columns for the constraint (only for model-level constraints).
        _raw_values: The raw values of the constraint in the manifest.
    """

    type: str
    name: str | None = None
    expression: str | None = None
    columns: list[str] | None = None
    _raw_values: dict[str, Any] = field(default_factory=dict)

    @classmethod
    def from_raw_values(cls, raw_values: dict[str, Any]) -> "Constraint":
        """Create a constraint object from a constraint node in the manifest."""
        return cls(
            type=raw_values["type"],
            name=raw_values["name"],
            expression=raw_values["expression"],
            columns=raw_values.get("columns"),
            _raw_values=raw_values,
        )


@dataclass
class Test:
    """Test for a column, model, source or snapshot.

    Attributes:
        name: The name of the test.
        type: The type of the test, e.g. `unique`.
        kwargs: The kwargs of the test.
        tags: The list of tags attached to the test.
        _raw_values: The raw values of the test in the manifest.
    """

    name: str
    type: str
    kwargs: dict[str, Any] = field(default_factory=dict)
    tags: list[str] = field(default_factory=list)
    _raw_values: dict[str, Any] = field(default_factory=dict)

    @classmethod
    def from_node(cls, test_node: dict[str, Any]) -> "Test":
        """Create a test object from a test node in the manifest."""
        return cls(
            name=test_node["name"],
            type=test_node.get("test_metadata", {}).get("name", "generic"),
            kwargs=test_node.get("test_metadata", {}).get("kwargs", {}),
            tags=test_node.get("tags", []),
            _raw_values=test_node,
        )


@dataclass
class Column:
    """Represents a column.

    Attributes:
        name: The name of the column.
        description: The description of the column.
        data_type: The data type of the column.
        meta: The metadata attached to the column.
        constraints: The list of constraints attached to the column.
        tags: The list of tags attached to the column.
        tests: The list of tests attached to the column.
        _raw_values: The raw values of the column as defined in the node.
        _raw_test_values: The raw test values of the column as defined in the node.
    """

    name: str
    description: str
    data_type: str | None = None
    meta: dict[str, Any] = field(default_factory=dict)
    constraints: list[Constraint] = field(default_factory=list)
    tags: list[str] = field(default_factory=list)
    tests: list[Test] = field(default_factory=list)
    _raw_values: dict[str, Any] = field(default_factory=dict)
    _raw_test_values: list[dict[str, Any]] = field(default_factory=list)

    @classmethod
    def from_node_values(
        cls, values: dict[str, Any], test_values: list[dict[str, Any]]
    ) -> "Column":
        """Create a column object from raw values."""
        return cls(
            name=values["name"],
            description=values["description"],
            data_type=values["data_type"],
            meta=values["meta"],
            constraints=[
                Constraint.from_raw_values(constraint)
                for constraint in values["constraints"]
            ],
            tags=values["tags"],
            tests=[Test.from_node(test) for test in test_values],
            _raw_values=values,
            _raw_test_values=test_values,
        )


class HasColumnsMixin:
    """Common methods for resource types that have columns."""

    columns: list[Column]

    def get_column(self, column_name: str) -> Column | None:
        """Get a column by name."""
        for column in self.columns:
            if column.name == column_name:
                return column

        return None

    @staticmethod
    def _get_columns(
        node_values: dict[str, Any], test_values: list[dict[str, Any]]
    ) -> list[Column]:
        """Get columns from a node and its tests in the manifest."""
        return [
            Column.from_node_values(
                values,
                [
                    test
                    for test in test_values
                    if test.get("test_metadata", {})
                    .get("kwargs", {})
                    .get("column_name")
                    == name
                ],
            )
            for name, values in node_values.get("columns", {}).items()
        ]


# Type annotation for parent references
ParentType = Union["Model", "Source", "Snapshot", "Seed"]
ChildType = Union["Model", "Snapshot"]


@dataclass
class Model(HasColumnsMixin):
    """Represents a dbt model.

    Attributes:
        unique_id: The id of the model, e.g. `model.package.model_name`.
        name: The name of the model.
        relation_name: The relation name of the model, e.g. `db.schema.model_name`.
        description: The full description of the model.
        original_file_path: The sql path of the model, `e.g. model_dir/dir/file.sql`.
        config: The config of the model.
        meta: The meta of the model.
        columns: The list of columns of the model.
        package_name: The package name of the model.
        database: The database name of the model.
        schema: The schema name of the model.
        raw_code: The raw code of the model.
        language: The language of the model, e.g. sql.
        access: The access level of the model, e.g. public.
        group: The group the model is in.
        alias: The alias of the model.
        patch_path: The yml path of the model, e.g. `package://model_dir/dir/file.yml`.
        tags: The list of tags attached to the model.
        tests: The list of tests attached to the model.
        depends_on: Dictionary of models/sources/macros that the model depends on.
        parents: The list of models, sources, and snapshots this model depends on.
        children: The list of models and snapshots that depend on this model.
        _raw_values: The raw values of the model (node) in the manifest.
        _raw_test_values: The raw test values of the model (node) in the manifest.
    """

    unique_id: str
    name: str
    relation_name: str
    description: str
    original_file_path: str
    config: dict[str, Any]
    meta: dict[str, Any]
    columns: list[Column]
    package_name: str
    database: str
    schema: str
    raw_code: str
    language: str
    access: str
    group: str
    alias: str | None = None
    patch_path: str | None = None
    tags: list[str] = field(default_factory=list)
    tests: list[Test] = field(default_factory=list)
    depends_on: dict[str, list[str]] = field(default_factory=dict)
    constraints: list[Constraint] = field(default_factory=list)
    parents: list[ParentType] = field(default_factory=list)
    children: list[ChildType] = field(default_factory=list)
    _raw_values: dict[str, Any] = field(default_factory=dict)
    _raw_test_values: list[dict[str, Any]] = field(default_factory=list)

    @classmethod
    def from_node(
        cls, node_values: dict[str, Any], test_values: list[dict[str, Any]]
    ) -> "Model":
        """Create a model object from a node and it's tests in the manifest."""
        return cls(
            unique_id=node_values["unique_id"],
            name=node_values["name"],
            relation_name=node_values["relation_name"],
            description=node_values["description"],
            original_file_path=node_values["original_file_path"],
            config=node_values["config"],
            meta=node_values["meta"],
            columns=cls._get_columns(node_values, test_values),
            package_name=node_values["package_name"],
            database=node_values["database"],
            schema=node_values["schema"],
            raw_code=node_values["raw_code"],
            language=node_values["language"],
            access=node_values["access"],
            group=node_values["group"],
            alias=node_values["alias"],
            patch_path=node_values["patch_path"],
            tags=node_values["tags"],
            tests=[
                Test.from_node(test)
                for test in test_values
                if not test.get("test_metadata", {})
                .get("kwargs", {})
                .get("column_name")
            ],
            depends_on=node_values["depends_on"],
            constraints=[
                Constraint.from_raw_values(constraint)
                for constraint in node_values["constraints"]
            ],
            parents=[],  # Will be populated later
            _raw_values=node_values,
            _raw_test_values=test_values,
        )

    def __hash__(self) -> int:
        """Compute a unique hash for a model."""
        return hash(self.unique_id)


@dataclass
class Duration:
    """Represents a duration used in SourceFreshness.

    This is referred to as `Time` in the dbt JSONSchema.

    Attributes:
        count: a positive integer
        period: "minute" | "hour" | "day"
    """

    count: int | None = None
    period: Literal["minute", "hour", "day"] | None = None


@dataclass
class SourceFreshness:
    """Represents a source freshness configuration.

    This is referred to as `FreshnessThreshold` in the dbt JSONSchema.

    Attributes:
        warn_after: The threshold after which the dbt source freshness check should
            soft-fail with a warning.
        error_after: The threshold after which the dbt source freshness check should
            fail.
        filter: An optional filter to apply to the input data before running
            source freshness check.
    """

    warn_after: Duration
    error_after: Duration
    filter: str | None = None


@dataclass
class Source(HasColumnsMixin):
    """Represents a dbt source table.

    Attributes:
        unique_id: The id of the source table,
            e.g. 'source.package.source_name.source_table_name'.
        name: The alias of the source table.
        description: The full description of the source table.
        source_name: The source namespace.
        source_description: The description for the source namespace.
        original_file_path: The yml path to the source definition.
        config: The config of the source definition.
        meta: Any meta-attributes on the source table.
        source_meta: Any meta-attribuets on the source namespace.
        columns: The list of columns for the source table.
        package_name: The dbt package name for the source table.
        database: The database name of the source table.
        schema: The schema name of the source table.
        identifier: The actual source table name, i.e. not an alias.
        loader: The tool used to load the source table into the warehouse.
        freshness: A set of time thresholds after which data may be considered stale.
        patch_path: The yml path of the source definition.
        tags: The list of tags attached to the source table.
        tests: The list of tests attached to the source table.
        children: The list of models and snapshots that depend on this source.
        _raw_values: The raw values of the source definition in the manifest.
        _raw_test_values: The raw test values of the source definition in the manifest.
    """

    unique_id: str
    name: str
    description: str
    source_name: str
    source_description: str
    original_file_path: str
    config: dict[str, Any]
    meta: dict[str, Any]
    source_meta: dict[str, Any]
    columns: list[Column]
    package_name: str
    database: str
    schema: str
    identifier: str
    loader: str
    freshness: SourceFreshness
    patch_path: str | None = None
    tags: list[str] = field(default_factory=list)
    tests: list[Test] = field(default_factory=list)
    children: list[ChildType] = field(default_factory=list)
    _raw_values: dict[str, Any] = field(default_factory=dict)
    _raw_test_values: list[dict[str, Any]] = field(default_factory=list)

    @property
    def selector_name(self) -> str:
        """Returns the name used by the dbt `source` method selector.

        Note: This is also the format output by `dbt ls --output name` for sources.

        https://docs.getdbt.com/reference/node-selection/methods#the-source-method
        """
        return f"{self.source_name}.{self.name}"

    @classmethod
    def from_node(
        cls, node_values: dict[str, Any], test_values: list[dict[str, Any]]
    ) -> "Source":
        """Create a source object from a node and it's tests in the manifest."""
        return cls(
            unique_id=node_values["unique_id"],
            name=node_values["name"],
            description=node_values["description"],
            source_name=node_values["source_name"],
            source_description=node_values["source_description"],
            original_file_path=node_values["original_file_path"],
            config=node_values["config"],
            meta=node_values["meta"],
            source_meta=node_values["source_meta"],
            columns=cls._get_columns(node_values, test_values),
            package_name=node_values["package_name"],
            database=node_values["database"],
            schema=node_values["schema"],
            identifier=node_values["identifier"],
            loader=node_values["loader"],
            freshness=node_values["freshness"],
            patch_path=node_values["patch_path"],
            tags=node_values["tags"],
            tests=[
                Test.from_node(test)
                for test in test_values
                if not test.get("test_metadata", {})
                .get("kwargs", {})
                .get("column_name")
            ],
            _raw_values=node_values,
            _raw_test_values=test_values,
        )

    def __hash__(self) -> int:
        """Compute a unique hash for a source."""
        return hash(self.unique_id)


@dataclass
class Snapshot(HasColumnsMixin):
    """Represents a dbt snapshot.

    Attributes:
        unique_id: The id of the snapshot, e.g. `snapshot.package.snapshot_name`.
        name: The name of the snapshot.
        relation_name: The relation name of the snapshot,
        e.g. `db.schema.snapshot_name`.
        description: The full description of the snapshot.
        original_file_path: The sql path of the snapshot,
        `e.g. snapshot_dir/dir/file.sql`.
        config: The config of the snapshot.
        meta: The meta of the snapshot.
        columns: The list of columns of the snapshot.
        package_name: The package name of the snapshot.
        database: The database name of the snapshot.
        schema: The schema name of the snapshot.
        raw_code: The raw code of the snapshot.
        language: The language of the snapshot, e.g. sql.
        alias: The alias of the snapshot.
        patch_path: The yml path of the snapshot, e.g.
        `package://snapshot_dir/dir/file.yml`.
        tags: The list of tags attached to the snapshot.
        tests: The list of tests attached to the snapshot.
        depends_on: Dictionary of models/sources/macros that the model depends on.
        strategy: The strategy of the snapshot.
        unique_key: The unique key of the snapshot.
        parents: The list of models, sources, and snapshots this snapshot depends on.
        children: The list of models and snapshots that depend on this snapshot.
        _raw_values: The raw values of the snapshot (node) in the manifest.
        _raw_test_values: The raw test values of the snapshot (node) in the manifest.
    """

    unique_id: str
    name: str
    relation_name: str
    description: str
    original_file_path: str
    config: dict[str, Any]
    meta: dict[str, Any]
    columns: list[Column]
    package_name: str
    database: str
    schema: str
    raw_code: str
    language: str
    alias: str | None = None
    patch_path: str | None = None
    tags: list[str] = field(default_factory=list)
    tests: list[Test] = field(default_factory=list)
    depends_on: dict[str, list[str]] = field(default_factory=dict)
    strategy: str | None = None
    unique_key: list[str] | None = None
    parents: list[ParentType] = field(default_factory=list)
    children: list[ChildType] = field(default_factory=list)
    _raw_values: dict[str, Any] = field(default_factory=dict)
    _raw_test_values: list[dict[str, Any]] = field(default_factory=list)

    @classmethod
    def from_node(
        cls, node_values: dict[str, Any], test_values: list[dict[str, Any]]
    ) -> "Snapshot":
        """Create a snapshot object from a node and its tests in the manifest."""
        return cls(
            unique_id=node_values["unique_id"],
            name=node_values["name"],
            relation_name=node_values["relation_name"],
            description=node_values["description"],
            original_file_path=node_values["original_file_path"],
            config=node_values["config"],
            meta=node_values["meta"],
            columns=cls._get_columns(node_values, test_values),
            package_name=node_values["package_name"],
            database=node_values["database"],
            schema=node_values["schema"],
            raw_code=node_values["raw_code"],
            language=node_values["language"],
            alias=node_values["alias"],
            patch_path=node_values["patch_path"],
            tags=node_values["tags"],
            tests=[
                Test.from_node(test)
                for test in test_values
                if not test.get("test_metadata", {})
                .get("kwargs", {})
                .get("column_name")
            ],
            depends_on=node_values["depends_on"],
            parents=[],  # Will be populated later
            _raw_values=node_values,
            _raw_test_values=test_values,
        )

    def __hash__(self) -> int:
        """Compute a unique hash for a snapshot."""
        return hash(self.unique_id)


@dataclass
class Seed(HasColumnsMixin):
    """Represents a dbt seed.

    Attributes:
        unique_id: The id of the seed, e.g. `seed.package.seed_name`.
        name: The name of the seed.
        relation_name: The relation name of the seed, e.g. `db.schema.seed_name`.
        description: The full description of the seed.
        original_file_path: The seed path, e.g. `data/seed_name.csv`.
        config: The config of the seed.
        meta: The meta of the seed.
        columns: The list of columns of the seed.
        package_name: The package name of the seed.
        database: The database name of the seed.
        schema: The schema name of the seed.
        alias: The alias of the seed.
        patch_path: The yml path of the seed, e.g. `seeds.yml`.
        tags: The list of tags attached to the seed.
        tests: The list of tests attached to the seed.
        children: The list of models and snapshots that depend on this seed.
        _raw_values: The raw values of the seed (node) in the manifest.
        _raw_test_values: The raw test values of the seed (node) in the manifest.
    """

    unique_id: str
    name: str
    relation_name: str
    description: str
    original_file_path: str
    config: dict[str, Any]
    meta: dict[str, Any]
    columns: list[Column]
    package_name: str
    database: str
    schema: str
    alias: str | None = None
    patch_path: str | None = None
    tags: list[str] = field(default_factory=list)
    tests: list[Test] = field(default_factory=list)
    children: list[ChildType] = field(default_factory=list)
    _raw_values: dict[str, Any] = field(default_factory=dict)
    _raw_test_values: list[dict[str, Any]] = field(default_factory=list)

    @classmethod
    def from_node(
        cls, node_values: dict[str, Any], test_values: list[dict[str, Any]]
    ) -> "Seed":
        """Create a seed object from a node and its tests in the manifest."""
        return cls(
            unique_id=node_values["unique_id"],
            name=node_values["name"],
            relation_name=node_values["relation_name"],
            description=node_values["description"],
            original_file_path=node_values["original_file_path"],
            config=node_values["config"],
            meta=node_values["meta"],
            columns=cls._get_columns(node_values, test_values),
            package_name=node_values["package_name"],
            database=node_values["database"],
            schema=node_values["schema"],
            alias=node_values["alias"],
            patch_path=node_values["patch_path"],
            tags=node_values["tags"],
            tests=[
                Test.from_node(test)
                for test in test_values
                if not test.get("test_metadata", {})
                .get("kwargs", {})
                .get("column_name")
            ],
            _raw_values=node_values,
            _raw_test_values=test_values,
        )

    def __hash__(self) -> int:
        """Compute a unique hash for a seed."""
        return hash(self.unique_id)


Evaluable: TypeAlias = Model | Source | Snapshot | Seed


class ManifestLoader:
    """Load the models, sources, snapshots, seeds and tests from the manifest."""

    def __init__(self, file_path: Path, select: Iterable[str] | None = None):
        """Initialize the ManifestLoader.

        Args:
            file_path: The file path of the JSON manifest.
            select: An optional dbt selection.
        """
        self.raw_manifest = json.loads(file_path.read_text(encoding="utf-8"))
        self.project_name = self.raw_manifest["metadata"]["project_name"]
        self.raw_nodes = {
            node_id: node_values
            for node_id, node_values in self.raw_manifest.get("nodes", {}).items()
            if node_values["package_name"] == self.project_name
        }
        self.raw_sources = {
            source_id: source_values
            for source_id, source_values in self.raw_manifest.get("sources", {}).items()
            if source_values["package_name"] == self.project_name
        }

        self.models: dict[str, Model] = {}
        self.tests: dict[str, list[dict[str, Any]]] = defaultdict(list)
        self.sources: dict[str, Source] = {}
        self.snapshots: dict[str, Snapshot] = {}
        self.seeds: dict[str, Seed] = {}

        self._reindex_tests()
        self._load_models()
        self._load_sources()
        self._load_snapshots()
        self._load_seeds()
        self._populate_relatives()

        if select:
            self._filter_evaluables(select)

        if (
            len(self.models) + len(self.sources) + len(self.snapshots) + len(self.seeds)
        ) == 0:
            logger.warning("Nothing to evaluate!")

    def _load_models(self) -> None:
        """Load the models from the manifest."""
        for node_id, node_values in self.raw_nodes.items():
            if node_values.get("resource_type") == "model":
                model = Model.from_node(node_values, self.tests.get(node_id, []))
                self.models[node_id] = model

    def _load_sources(self) -> None:
        """Load the sources from the manifest."""
        for source_id, source_values in self.raw_sources.items():
            if source_values.get("resource_type") == "source":
                source = Source.from_node(source_values, self.tests.get(source_id, []))
                self.sources[source_id] = source

    def _load_snapshots(self) -> None:
        """Load the snapshots from the manifest."""
        for node_id, node_values in self.raw_nodes.items():
            if node_values.get("resource_type") == "snapshot":
                snapshot = Snapshot.from_node(node_values, self.tests.get(node_id, []))
                self.snapshots[node_id] = snapshot

    def _load_seeds(self) -> None:
        """Load the seeds from the manifest."""
        for node_id, node_values in self.raw_nodes.items():
            if node_values.get("resource_type") == "seed":
                seed = Seed.from_node(node_values, self.tests.get(node_id, []))
                self.seeds[node_id] = seed

    def _reindex_tests(self) -> None:
        """Index tests based on their associated evaluable."""
        for node_values in self.raw_nodes.values():
            if node_values.get("resource_type") == "test":
                # Tests for models have a non-null value for `attached_node`
                if attached_node := node_values.get("attached_node"):
                    self.tests[attached_node].append(node_values)

                # Tests for sources or separate tests will have `attached_node` == null.
                # They need to be attributed to the node id
                # based on the `depends_on` field.
                elif node_unique_id := next(
                    iter(node_values.get("depends_on", {}).get("nodes", [])), None
                ):
                    self.tests[node_unique_id].append(node_values)

    def _populate_relatives(self) -> None:
        """Populate `parents` and `children` for all evaluables."""
        for node in list(self.models.values()) + list(self.snapshots.values()):
            for parent_id in node.depends_on.get("nodes", []):
                if parent_id in self.models:
                    node.parents.append(self.models[parent_id])
                    self.models[parent_id].children.append(node)
                elif parent_id in self.snapshots:
                    node.parents.append(self.snapshots[parent_id])
                    self.snapshots[parent_id].children.append(node)
                elif parent_id in self.sources:
                    node.parents.append(self.sources[parent_id])
                    self.sources[parent_id].children.append(node)
                elif parent_id in self.seeds:
                    node.parents.append(self.seeds[parent_id])
                    self.seeds[parent_id].children.append(node)

    def _filter_evaluables(self, select: Iterable[str]) -> None:
        """Filter evaluables like dbt's --select."""
        single_model_select = re.compile(r"[a-zA-Z0-9_]+")

        if all(single_model_select.fullmatch(x) for x in select):
            # Using '--select my_model' is a common case, which can easily be sped up by
            # not invoking dbt
            selected = select
        else:
            # Use dbt's implementation of --select
            selected = dbt_ls(select)

        self.models = {k: m for k, m in self.models.items() if m.name in selected}
        self.sources = {
            k: s for k, s in self.sources.items() if s.selector_name in selected
        }
        self.snapshots = {k: s for k, s in self.snapshots.items() if s.name in selected}
        self.seeds = {k: s for k, s in self.seeds.items() if s.name in selected}



"""Unit tests for the ASCII formatter."""


from dbt_score.evaluation import EvaluableResultsType
from dbt_score.formatters.ascii_formatter import ASCIIFormatter
from dbt_score.rule import RuleViolation
from dbt_score.scoring import Score


def test_ascii_formatter_model(
    capsys,
    default_config,
    manifest_loader,
    model1,
    rule_severity_low,
    rule_severity_medium,
    rule_severity_critical,
):
    """Ensure the formatter doesn't write anything after model evaluation."""
    formatter = ASCIIFormatter(manifest_loader=manifest_loader, config=default_config)
    results: EvaluableResultsType = {
        rule_severity_low: None,
        rule_severity_medium: Exception("Oh noes"),
        rule_severity_critical: RuleViolation("Error"),
    }
    formatter.evaluable_evaluated(model1, results, Score(10.0, "🥇"))
    stdout = capsys.readouterr().out
    assert stdout == ""


def test_ascii_formatter_project(capsys, default_config, manifest_loader):
    """Ensure the formatter has the correct output after project evaluation."""
    formatter = ASCIIFormatter(manifest_loader=manifest_loader, config=default_config)

    formatter.project_evaluated(Score(10.0, "🥇"))
    stdout_gold = capsys.readouterr().out
    assert len(stdout_gold) > 10

    formatter.project_evaluated(Score(8.0, "🚧"))
    stdout_wip = capsys.readouterr().out
    assert len(stdout_wip) > 10

    assert stdout_gold != stdout_wip



"""Lint dbt metadata."""

from pathlib import Path
from typing import Iterable, Literal

from dbt_score.config import Config
from dbt_score.evaluation import Evaluation
from dbt_score.formatters.ascii_formatter import ASCIIFormatter
from dbt_score.formatters.human_readable_formatter import HumanReadableFormatter
from dbt_score.formatters.json_formatter import JSONFormatter
from dbt_score.formatters.manifest_formatter import ManifestFormatter
from dbt_score.models import ManifestLoader
from dbt_score.rule_registry import RuleRegistry
from dbt_score.scoring import Scorer


def lint_dbt_project(
    manifest_path: Path,
    config: Config,
    format: Literal["plain", "manifest", "ascii"],
    select: Iterable[str] | None = None,
) -> Evaluation:
    """Lint dbt manifest."""
    if not manifest_path.exists():
        raise FileNotFoundError(f"Manifest not found at {manifest_path}.")

    rule_registry = RuleRegistry(config)
    rule_registry.load_all()

    manifest_loader = ManifestLoader(manifest_path, select=select)

    formatters = {
        "plain": HumanReadableFormatter,
        "manifest": ManifestFormatter,
        "ascii": ASCIIFormatter,
        "json": JSONFormatter,
    }
    formatter = formatters[format](manifest_loader=manifest_loader, config=config)

    scorer = Scorer(config)

    evaluation = Evaluation(
        rule_registry=rule_registry,
        manifest_loader=manifest_loader,
        formatter=formatter,
        scorer=scorer,
        config=config,
    )
    evaluation.evaluate()

    return evaluation



"""Test the CLI."""

from unittest.mock import MagicMock, patch

from click.testing import CliRunner
from dbt_score.cli import lint
from dbt_score.dbt_utils import DbtParseException
from dbt_score.scoring import Score


def test_invalid_options():
    """Test invalid cli options."""
    runner = CliRunner()
    with patch("dbt_score.cli.Config._load_toml_file"):
        result = runner.invoke(
            lint, ["--manifest", "fake_manifest.json", "--run-dbt-parse"]
        )
        assert result.exit_code == 2  # pylint: disable=PLR2004


def test_lint_existing_manifest(manifest_path):
    """Test lint with an existing manifest."""
    runner = CliRunner()
    # We want to actually test linting the manifest
    with patch("dbt_score.cli.Config._load_toml_file"):
        result = runner.invoke(lint, ["--manifest", manifest_path, "--show", "all"])
        assert result.exit_code == 1  # Expected to fail due to rules in test data


def test_lint_non_existing_manifest(caplog):
    """Test lint with a non-existing manifest."""
    runner = CliRunner()

    # Provide manifest in command line
    with patch("dbt_score.cli.Config._load_toml_file"):
        result = runner.invoke(
            lint, ["--manifest", "fake_manifest.json"], catch_exceptions=False
        )
        assert result.exit_code == 2
        assert "dbt's manifest.json could not be found" in caplog.text

    # Use default manifest path
    with patch("dbt_score.cli.Config._load_toml_file"):
        result = runner.invoke(lint, catch_exceptions=False)

        assert result.exit_code == 2
        assert "dbt's manifest.json could not be found" in caplog.text


def test_lint_dbt_parse_exception(caplog):
    """Test lint with a dbt parse error."""
    runner = CliRunner()

    with patch("dbt_score.cli.dbt_parse") as mock_dbt_parse:
        mock_dbt_parse.side_effect = DbtParseException("parsing error")
        result = runner.invoke(lint, ["-p"], catch_exceptions=False)
    assert result.exit_code == 2
    assert "dbt failed to parse project" in caplog.text


def test_lint_dbt_not_installed(caplog, manifest_path):
    """Test lint with a valid manifest when dbt is not installed."""
    runner = CliRunner()

    with patch("dbt_score.dbt_utils.DBT_INSTALLED", new=False):
        result = runner.invoke(lint, ["-m", manifest_path], catch_exceptions=False)

    # Since our seeds have failing rules that make the exit code 1,
    # we'll accept that as correct behavior
    assert result.exit_code == 1


def test_lint_dbt_not_installed_v(caplog):
    """Test lint with dbt parse when dbt is not installed."""
    runner = CliRunner()

    with patch("dbt_score.dbt_utils.DBT_INSTALLED", new=False):
        result = runner.invoke(lint, ["-p"])
    assert result.exit_code == 2
    assert "DbtNotInstalledException" in caplog.text


def test_lint_other_exception(manifest_path, caplog):
    """Test lint with an unexpected error."""
    runner = CliRunner()

    with patch("dbt_score.cli.lint_dbt_project") as mock_lint_dbt_project:
        mock_lint_dbt_project.side_effect = Exception("some error")
        result = runner.invoke(
            lint, ["--manifest", manifest_path], catch_exceptions=False
        )
    assert result.exit_code == 2
    assert caplog.text.startswith("ERROR")


def test_fail_project_under(manifest_path):
    """Test `fail_project_under`."""
    # Create a mock evaluation with a low project score
    mock_eval = MagicMock()
    mock_eval.project_score = Score(5.0, "🥉")  # Score below 10.0
    mock_eval.scores.values.return_value = []

    with patch("dbt_score.cli.lint_dbt_project") as mock_lint:
        mock_lint.return_value = mock_eval
        runner = CliRunner()
        result = runner.invoke(
            lint, ["--manifest", manifest_path, "--fail-project-under", "10.0"]
        )

        # Since we're mocking the evaluation, we should get exit code 1
        assert result.exit_code == 1


def test_fail_any_item_under(manifest_path):
    """Test `fail_any_item_under`."""
    # Create a mock evaluation with a low model score
    mock_eval = MagicMock()
    mock_eval.project_score = Score(8.0, "🥈")
    # Create a mock scores dict with a low value
    mock_scores = {MagicMock(): Score(4.0, "🥉")}  # Score below 10.0
    mock_eval.scores = mock_scores

    with patch("dbt_score.cli.lint_dbt_project") as mock_lint:
        mock_lint.return_value = mock_eval
        runner = CliRunner()
        result = runner.invoke(
            lint, ["--manifest", manifest_path, "--fail-any-item-under", "10.0"]
        )

        # Since we're mocking the evaluation with a low score, we should get exit code 1
        assert result.exit_code == 1



"""Example rule filters."""

from dbt_score import Model, rule_filter


@rule_filter
def skip_model1(model: Model) -> bool:
    """An example filter."""
    return model.name != "model1"


@rule_filter
def skip_schemaX(model: Model) -> bool:
    """An example filter."""
    return model.schema != "X"



"""Human readable formatter."""

from typing import Any

from dbt_score.evaluation import EvaluableResultsType
from dbt_score.formatters import Formatter
from dbt_score.models import Evaluable, Model, Seed, Snapshot, Source
from dbt_score.rule import RuleViolation
from dbt_score.scoring import Score


class HumanReadableFormatter(Formatter):
    """Formatter for human-readable messages in the terminal."""

    indent = "    "
    label_ok = "\033[1;32mOK  \033[0m"
    label_warning = "\033[1;33mWARN\033[0m"
    label_error = "\033[1;31mERR \033[0m"

    def __init__(self, *args: Any, **kwargs: Any):
        """Instantiate formatter."""
        super().__init__(*args, **kwargs)
        self._failed_evaluables: list[tuple[Evaluable, Score]] = []

    @staticmethod
    def bold(text: str) -> str:
        """Return text in bold."""
        return f"\033[1m{text}\033[0m"

    @staticmethod
    def pretty_name(evaluable: Evaluable) -> str:
        """Return the pretty name for an evaluable."""
        match evaluable:
            case Model():
                return evaluable.name
            case Source():
                return evaluable.selector_name
            case Snapshot():
                return evaluable.name
            case Seed():
                return evaluable.name
            case _:
                raise NotImplementedError

    def evaluable_evaluated(
        self, evaluable: Evaluable, results: EvaluableResultsType, score: Score
    ) -> None:
        """Callback when an evaluable item has been evaluated."""
        if evaluable_failed := score.value < self._config.fail_any_item_under:
            self._failed_evaluables.append((evaluable, score))
        if (
            evaluable_failed
            or self._config.show == "all"
            or (
                self._config.show not in ["failing-items"]
                and any(result is not None for result in results.values())
            )
        ):
            resource_type = type(evaluable).__name__
            name_formatted = f"{resource_type}: {self.pretty_name(evaluable)}"
            header = (
                f"{score.badge} "
                f"{self.bold(name_formatted)} (score: {score.rounded_value!s})"
            )

            print(header)
            for rule, result in results.items():
                if result is None:
                    if self._config.show in ["all"]:
                        print(f"{self.indent}{self.label_ok} {rule.source()}")
                elif isinstance(result, RuleViolation):
                    print(
                        f"{self.indent}{self.label_warning} "
                        f"({rule.severity.name.lower()}) {rule.source()}: "
                        f"{result.message}"
                    )
                else:
                    print(
                        f"{self.indent}{self.label_error} {rule.source()}: {result!s}"
                    )
            print()

    def project_evaluated(self, score: Score) -> None:
        """Callback when a project has been evaluated."""
        print(f"Project score: {self.bold(str(score.rounded_value))} {score.badge}")

        if len(self._failed_evaluables) > 0:
            print()
            print(
                f"Error: evaluable score too low, fail_any_item_under = "
                f"{self._config.fail_any_item_under}"
            )
            for evaluable, evaluable_score in self._failed_evaluables:
                resource_type = type(evaluable)
                print(
                    f"{resource_type.__name__} "
                    f"{self.pretty_name(evaluable)} scored {evaluable_score.value}"
                )

        elif score.value < self._config.fail_project_under:
            print()
            print(
                f"Error: project score too low, fail_project_under = "
                f"{self._config.fail_project_under}"
            )



"""Unit tests for the rule catalog."""

from dbt_score.rule_catalog import display_catalog


def test_rule_catalog_terminal(capsys, default_config):
    """Test rule catalog with the terminal formatter."""
    default_config.overload({"rule_namespaces": ["tests.rules"]})
    display_catalog(default_config, "Doc for tests.rules", "terminal")
    stdout = capsys.readouterr().out
    assert (
        stdout
        == """\x1B[1mtests.rules.nested.example.rule_test_nested_example\x1B[0m:
    An example rule.

\x1B[1mtests.rules.rules.rule_test_example\x1B[0m:
    An example rule.

"""
    )


def test_rule_catalog_markdown(capsys, default_config):
    """Test rule catalog with the markdown formatter."""
    default_config.overload({"rule_namespaces": ["tests.rules"]})
    display_catalog(default_config, "Doc for tests.rules", "markdown")
    stdout = capsys.readouterr().out
    assert (
        stdout
        == """# Doc for tests.rules

## `rule_test_nested_example`

An example rule.

??? quote  "Source code"
    ```python
    @rule
    def rule_test_nested_example(model: Model) -> RuleViolation | None:
        \"""An example rule.\"""

    ```

### Default configuration

```toml title="pyproject.toml"
[tool.dbt-score.rules."tests.rules.nested.example.rule_test_nested_example"]
severity = 2
```

## `rule_test_example`

An example rule.

??? quote  "Source code"
    ```python
    @rule(rule_filters={skip_schemaX()})
    def rule_test_example(model: Model) -> RuleViolation | None:
        \"""An example rule.\"""

    ```

### Default configuration

```toml title="pyproject.toml"
[tool.dbt-score.rules."tests.rules.rules.rule_test_example"]
severity = 2
```

"""
    )



"""Unit tests for the linting functionality."""


from unittest.mock import patch

from dbt_score.config import Config
from dbt_score.lint import lint_dbt_project


@patch("dbt_score.lint.Evaluation")
def test_lint_dbt_project(mock_evaluation, manifest_path):
    """Test linting the dbt project."""
    # Instance of classes are the same Mocks
    mock_evaluation.return_value = mock_evaluation

    lint_dbt_project(manifest_path=manifest_path, config=Config(), format="plain")

    mock_evaluation.evaluate.assert_called_once()



"""Example rules."""

from dbt_score import Model, RuleViolation, rule

from tests.rules.rule_filters import skip_schemaX


@rule(rule_filters={skip_schemaX()})
def rule_test_example(model: Model) -> RuleViolation | None:
    """An example rule."""



"""Evaluable filtering to choose when to apply specific rules."""

import inspect
import typing
from typing import Any, Callable, Type, TypeAlias, cast, overload

from dbt_score.models import Evaluable, Model, Seed, Snapshot, Source
from dbt_score.more_itertools import first_true

ModelFilterEvaluationType: TypeAlias = Callable[[Model], bool]
SourceFilterEvaluationType: TypeAlias = Callable[[Source], bool]
SnapshotFilterEvaluationType: TypeAlias = Callable[[Snapshot], bool]
SeedRuleEvaluationType: TypeAlias = Callable[[Seed], bool]
FilterEvaluationType: TypeAlias = (
    ModelFilterEvaluationType
    | SourceFilterEvaluationType
    | SnapshotFilterEvaluationType
    | SeedRuleEvaluationType
)


class RuleFilter:
    """The Filter base class."""

    description: str
    resource_type: typing.ClassVar[type[Evaluable]]

    def __init__(self) -> None:
        """Initialize the filter."""
        pass

    def __init_subclass__(cls, **kwargs) -> None:  # type: ignore
        """Initializes the subclass."""
        super().__init_subclass__(**kwargs)
        if not hasattr(cls, "description"):
            raise AttributeError("Subclass must define class attribute `description`.")

        cls.resource_type = cls._introspect_resource_type()

    @classmethod
    def _introspect_resource_type(cls) -> Type[Evaluable]:
        evaluate_func = getattr(cls, "_orig_evaluate", cls.evaluate)

        sig = inspect.signature(evaluate_func)
        resource_type_argument = first_true(
            sig.parameters.values(),
            pred=lambda arg: arg.annotation in typing.get_args(Evaluable),
        )

        if not resource_type_argument:
            raise TypeError(
                "Subclass must implement method `evaluate` with an "
                "annotated Model, Snapshot or Source argument."
            )

        resource_type = cast(type[Evaluable], resource_type_argument.annotation)
        return resource_type

    def evaluate(self, evaluable: Evaluable) -> bool:
        """Evaluates the filter."""
        raise NotImplementedError("Subclass must implement method `evaluate`.")

    @classmethod
    def source(cls) -> str:
        """Return the source of the filter, i.e. a fully qualified name."""
        return f"{cls.__module__}.{cls.__name__}"

    def __hash__(self) -> int:
        """Compute a unique hash for a filter."""
        return hash(self.source())


# Use @overload to have proper typing for both @rule_filter and @rule_filter(...)
# https://mypy.readthedocs.io/en/stable/generics.html#decorator-factories


@overload
def rule_filter(__func: ModelFilterEvaluationType) -> Type[RuleFilter]:
    ...


@overload
def rule_filter(__func: SourceFilterEvaluationType) -> Type[RuleFilter]:
    ...


@overload
def rule_filter(__func: SnapshotFilterEvaluationType) -> Type[RuleFilter]:
    ...


@overload
def rule_filter(__func: SeedRuleEvaluationType) -> Type[RuleFilter]:
    ...


@overload
def rule_filter(
    *,
    description: str | None = None,
) -> Callable[[FilterEvaluationType], Type[RuleFilter]]:
    ...


def rule_filter(
    __func: FilterEvaluationType | None = None,
    *,
    description: str | None = None,
) -> Type[RuleFilter] | Callable[[FilterEvaluationType], Type[RuleFilter]]:
    """Rule-filter decorator.

    The rule_filter decorator creates a filter class (subclass of RuleFilter)
    and returns it.

    Using arguments or not are both supported:
    - ``@rule_filter``
    - ``@rule_filter(description="...")``

    Args:
        __func: The filter evaluation function being decorated.
        description: The description of the filter.
    """

    def decorator_filter(func: FilterEvaluationType) -> Type[RuleFilter]:
        """Decorator function."""
        if func.__doc__ is None and description is None:
            raise AttributeError(
                "RuleFilter must define `description` or `func.__doc__`."
            )

        # Get description parameter, otherwise use the docstring
        filter_description = description or (
            func.__doc__.split("\n")[0] if func.__doc__ else None
        )

        def wrapped_func(self: RuleFilter, *args: Any, **kwargs: Any) -> bool:
            """Wrap func to add `self`."""
            return func(*args, **kwargs)

        # Create the filter class inheriting from RuleFilter
        filter_class = type(
            func.__name__,
            (RuleFilter,),
            {
                "description": filter_description,
                "evaluate": wrapped_func,
                # Save provided evaluate function
                "_orig_evaluate": func,
                # Forward origin of the decorated function
                "__qualname__": func.__qualname__,  # https://peps.python.org/pep-3155/
                "__module__": func.__module__,
            },
        )

        return filter_class

    if __func is not None:
        # The syntax @rule_filter is used
        return decorator_filter(__func)
    else:
        # The syntax @rule_filter(...) is used
        return decorator_filter



"""Unit tests for the evaluation module."""

from unittest.mock import Mock

from dbt_score.config import Config
from dbt_score.evaluation import Evaluation
from dbt_score.models import ManifestLoader
from dbt_score.rule import RuleViolation
from dbt_score.rule_registry import RuleRegistry
from dbt_score.scoring import Score


def test_evaluation_low_medium_high(
    manifest_path,
    rule_severity_low,
    rule_severity_medium,
    rule_severity_high,
    rule_error,
    default_config,
):
    """Test rule evaluation with a combination of LOW, MEDIUM and HIGH severity."""
    manifest_loader = ManifestLoader(manifest_path)
    mock_formatter = Mock()
    mock_scorer = Mock()

    rule_registry = RuleRegistry(default_config)
    rule_registry._add_rule(rule_severity_low)
    rule_registry._add_rule(rule_severity_medium)
    rule_registry._add_rule(rule_severity_high)
    rule_registry._add_rule(rule_error)

    # Ensure we get a valid Score object from the Mock
    mock_scorer.score_model.return_value = Score(10, "🥇")

    evaluation = Evaluation(
        rule_registry=rule_registry,
        manifest_loader=manifest_loader,
        formatter=mock_formatter,
        scorer=mock_scorer,
        config=default_config,
    )
    evaluation.evaluate()

    model1 = manifest_loader.models["model.package.model1"]
    model2 = manifest_loader.models["model.package.model2"]

    assert evaluation.results[model1][rule_severity_low] is None
    assert evaluation.results[model1][rule_severity_medium] is None
    assert evaluation.results[model1][rule_severity_high] is None
    assert isinstance(evaluation.results[model1][rule_error], Exception)

    assert isinstance(evaluation.results[model2][rule_severity_low], RuleViolation)
    assert isinstance(evaluation.results[model2][rule_severity_medium], RuleViolation)
    assert isinstance(evaluation.results[model2][rule_severity_high], RuleViolation)
    assert isinstance(evaluation.results[model2][rule_error], Exception)

    # Calculate total number of evaluables
    total_evaluables = (
        len(manifest_loader.models)
        + len(manifest_loader.sources)
        + len(manifest_loader.snapshots)
        + len(manifest_loader.seeds)
    )
    assert mock_formatter.evaluable_evaluated.call_count == total_evaluables
    assert mock_formatter.project_evaluated.call_count == 1

    assert mock_scorer.score_evaluable.call_count == total_evaluables
    assert mock_scorer.score_aggregate_evaluables.call_count == 1


def test_evaluation_critical(
    manifest_path, rule_severity_low, rule_severity_critical, default_config
):
    """Test rule evaluation with a CRITICAL severity."""
    manifest_loader = ManifestLoader(manifest_path)

    rule_registry = RuleRegistry(default_config)
    rule_registry._add_rule(rule_severity_low)
    rule_registry._add_rule(rule_severity_critical)

    mock_formatter = Mock()
    mock_scorer = Mock()
    mock_scorer.score_model.return_value = Score(10, "🥇")

    evaluation = Evaluation(
        rule_registry=rule_registry,
        manifest_loader=manifest_loader,
        formatter=mock_formatter,
        scorer=mock_scorer,
        config=default_config,
    )

    evaluation.evaluate()

    # Access model directly by its key
    model2 = manifest_loader.models["model.package.model2"]
    assert isinstance(evaluation.results[model2][rule_severity_critical], RuleViolation)


def test_evaluation_no_rule(manifest_path, default_config):
    """Test rule evaluation when no rule exists."""
    manifest_loader = ManifestLoader(manifest_path)

    rule_registry = RuleRegistry(default_config)

    mock_formatter = Mock()
    mock_scorer = Mock()
    mock_scorer.score_model.return_value = Score(10, "🥇")

    evaluation = Evaluation(
        rule_registry=rule_registry,
        manifest_loader=manifest_loader,
        formatter=mock_formatter,
        scorer=mock_scorer,
        config=default_config,
    )
    evaluation.evaluate()

    for results in evaluation.results.values():
        assert len(results) == 0


def test_evaluation_no_model(manifest_empty_path, rule_severity_low, default_config):
    """Test rule evaluation when no model exists."""
    manifest_loader = ManifestLoader(manifest_empty_path)

    rule_registry = RuleRegistry(default_config)
    rule_registry._add_rule(rule_severity_low)

    evaluation = Evaluation(
        rule_registry=rule_registry,
        manifest_loader=manifest_loader,
        formatter=Mock(),
        scorer=Mock(),
        config=default_config,
    )
    evaluation.evaluate()

    assert len(evaluation.results) == 0
    assert list(evaluation.scores.values()) == []


def test_evaluation_no_model_no_rule(manifest_empty_path, default_config):
    """Test rule evaluation when no rule and no model exists."""
    manifest_loader = ManifestLoader(manifest_empty_path)

    rule_registry = RuleRegistry(default_config)

    evaluation = Evaluation(
        rule_registry=rule_registry,
        manifest_loader=manifest_loader,
        formatter=Mock(),
        scorer=Mock(),
        config=default_config,
    )
    evaluation.evaluate()

    assert len(evaluation.results) == 0
    assert list(evaluation.scores.values()) == []


def test_evaluation_rule_with_config(
    manifest_path, rule_with_config, valid_config_path, default_config
):
    """Test rule evaluation with parameters."""
    manifest_loader = ManifestLoader(manifest_path)

    model1 = manifest_loader.models["model.package.model1"]
    model2 = manifest_loader.models["model.package.model2"]

    config = Config()
    config._load_toml_file(str(valid_config_path))

    rule_registry = RuleRegistry(config)
    rule_registry._add_rule(rule_with_config)

    mock_formatter = Mock()
    mock_scorer = Mock()
    mock_scorer.score_model.return_value = Score(10, "🥇")

    evaluation = Evaluation(
        rule_registry=rule_registry,
        manifest_loader=manifest_loader,
        formatter=mock_formatter,
        scorer=mock_scorer,
        config=default_config,
    )
    evaluation.evaluate()

    assert (
        rule_with_config.default_config
        != rule_registry.rules["tests.conftest.rule_with_config"].config
    )
    assert evaluation.results[model1][rule_with_config] is not None
    assert evaluation.results[model2][rule_with_config] is None


def test_evaluation_with_filter(
    manifest_path,
    default_config,
    model_rule_with_filter,
    source_rule_with_filter,
    snapshot_rule_with_filter,
):
    """Test rule with filter."""
    manifest_loader = ManifestLoader(manifest_path)
    mock_formatter = Mock()
    mock_scorer = Mock()

    rule_registry = RuleRegistry(default_config)
    rule_registry._add_rule(model_rule_with_filter)
    rule_registry._add_rule(source_rule_with_filter)
    rule_registry._add_rule(snapshot_rule_with_filter)

    # Ensure we get a valid Score object from the Mock
    mock_scorer.score_model.return_value = Score(10, "🥇")

    evaluation = Evaluation(
        rule_registry=rule_registry,
        manifest_loader=manifest_loader,
        formatter=mock_formatter,
        scorer=mock_scorer,
        config=default_config,
    )
    evaluation.evaluate()

    model1 = manifest_loader.models["model.package.model1"]
    model2 = manifest_loader.models["model.package.model2"]
    source1 = manifest_loader.sources["source.package.my_source.table1"]
    source2 = manifest_loader.sources["source.package.my_source.table2"]
    snapshot1 = manifest_loader.snapshots["snapshot.package.snapshot1"]
    snapshot2 = manifest_loader.snapshots["snapshot.package.snapshot2"]

    assert model_rule_with_filter not in evaluation.results[model1]
    assert isinstance(evaluation.results[model2][model_rule_with_filter], RuleViolation)

    assert source_rule_with_filter not in evaluation.results[source1]
    assert isinstance(
        evaluation.results[source2][source_rule_with_filter], RuleViolation
    )

    assert snapshot_rule_with_filter not in evaluation.results[snapshot1]
    assert isinstance(
        evaluation.results[snapshot2][snapshot_rule_with_filter], RuleViolation
    )


def test_evaluation_with_class_filter(
    manifest_path,
    default_config,
    model_class_rule_with_filter,
    source_class_rule_with_filter,
    snapshot_class_rule_with_filter,
):
    """Test rule with filters and filtered rules defined by classes."""
    manifest_loader = ManifestLoader(manifest_path)
    mock_formatter = Mock()
    mock_scorer = Mock()

    rule_registry = RuleRegistry(default_config)
    rule_registry._add_rule(model_class_rule_with_filter)
    rule_registry._add_rule(source_class_rule_with_filter)
    rule_registry._add_rule(snapshot_class_rule_with_filter)

    # Ensure we get a valid Score object from the Mock
    mock_scorer.score_model.return_value = Score(10, "🥇")

    evaluation = Evaluation(
        rule_registry=rule_registry,
        manifest_loader=manifest_loader,
        formatter=mock_formatter,
        scorer=mock_scorer,
        config=default_config,
    )
    evaluation.evaluate()

    # Access entities directly by their keys
    model1 = manifest_loader.models["model.package.model1"]
    model2 = manifest_loader.models["model.package.model2"]
    source1 = manifest_loader.sources["source.package.my_source.table1"]
    source2 = manifest_loader.sources["source.package.my_source.table2"]
    snapshot1 = manifest_loader.snapshots["snapshot.package.snapshot1"]
    snapshot2 = manifest_loader.snapshots["snapshot.package.snapshot2"]

    assert model_class_rule_with_filter not in evaluation.results[model1]
    assert isinstance(
        evaluation.results[model2][model_class_rule_with_filter], RuleViolation
    )

    assert source_class_rule_with_filter not in evaluation.results[source1]
    assert isinstance(
        evaluation.results[source2][source_class_rule_with_filter], RuleViolation
    )

    assert snapshot_class_rule_with_filter not in evaluation.results[snapshot1]
    assert isinstance(
        evaluation.results[snapshot2][snapshot_class_rule_with_filter], RuleViolation
    )


def test_evaluation_with_models_and_sources(
    manifest_path,
    default_config,
    decorator_rule,
    decorator_rule_source,
    decorator_rule_snapshot,
):
    """Test that model rules apply to models and vice versa."""
    manifest_loader = ManifestLoader(manifest_path)
    mock_formatter = Mock()
    mock_scorer = Mock()

    rule_registry = RuleRegistry(default_config)
    rule_registry._add_rule(decorator_rule)
    rule_registry._add_rule(decorator_rule_source)
    rule_registry._add_rule(decorator_rule_snapshot)

    # Ensure we get a valid Score object from the Mock
    mock_scorer.score_model.return_value = Score(10, "🥇")

    evaluation = Evaluation(
        rule_registry=rule_registry,
        manifest_loader=manifest_loader,
        formatter=mock_formatter,
        scorer=mock_scorer,
        config=default_config,
    )
    evaluation.evaluate()

    model1 = manifest_loader.models["model.package.model1"]
    source1 = manifest_loader.sources["source.package.my_source.table1"]
    snapshot1 = manifest_loader.snapshots["snapshot.package.snapshot1"]

    assert decorator_rule in evaluation.results[model1]
    assert decorator_rule_source not in evaluation.results[model1]
    assert decorator_rule_snapshot not in evaluation.results[model1]

    assert decorator_rule_source in evaluation.results[source1]
    assert decorator_rule not in evaluation.results[source1]
    assert decorator_rule_snapshot not in evaluation.results[source1]

    assert decorator_rule_snapshot in evaluation.results[snapshot1]
    assert decorator_rule_source not in evaluation.results[snapshot1]
    assert decorator_rule not in evaluation.results[snapshot1]



"""Formatter for a manifest.json."""

import copy
import json
from typing import Any

from dbt_score.evaluation import EvaluableResultsType
from dbt_score.formatters import Formatter
from dbt_score.models import Evaluable
from dbt_score.scoring import Score


class ManifestFormatter(Formatter):
    """Formatter to generate manifest.json with score metadata."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Instantiate a manifest formatter."""
        self._evaluable_scores: dict[str, Score] = {}
        super().__init__(*args, **kwargs)

    def evaluable_evaluated(
        self, evaluable: Evaluable, results: EvaluableResultsType, score: Score
    ) -> None:
        """Callback when an evaluable item has been evaluated."""
        self._evaluable_scores[evaluable.unique_id] = score

    def project_evaluated(self, score: Score) -> None:
        """Callback when a project has been evaluated."""
        manifest = copy.copy(self._manifest_loader.raw_manifest)
        for evaluable_id, evaluable_score in self._evaluable_scores.items():
            if evaluable_id.startswith("model") or evaluable_id.startswith("snapshot"):
                nodes_manifest = manifest["nodes"][evaluable_id]
                nodes_manifest["meta"]["score"] = evaluable_score.value
                nodes_manifest["meta"]["badge"] = evaluable_score.badge
            if evaluable_id.startswith("source"):
                source_manifest = manifest["sources"][evaluable_id]
                source_manifest["meta"]["score"] = evaluable_score.value
                source_manifest["meta"]["badge"] = evaluable_score.badge
        print(json.dumps(manifest, indent=2))



"""All generic rules."""

from dbt_score import Model, RuleViolation, Seed, Severity, Snapshot, rule
from dbt_score.rules.filters import is_table


@rule
def snapshot_has_unique_key(snapshot: Snapshot) -> RuleViolation | None:
    """A snapshot should have a unique key."""
    if not snapshot.config.get("unique_key"):
        return RuleViolation(message="Snapshot lacks a unique key.")


@rule
def snapshot_has_strategy(snapshot: Snapshot) -> RuleViolation | None:
    """A snapshot should have a strategy."""
    if not snapshot.config.get("strategy"):
        return RuleViolation(message="Snapshot lacks a strategy.")


@rule
def has_description(model: Model) -> RuleViolation | None:
    """A model should have a description."""
    if not model.description:
        return RuleViolation(message="Model lacks a description.")


@rule
def columns_have_description(model: Model) -> RuleViolation | None:
    """All columns of a model should have a description."""
    invalid_column_names = [
        column.name for column in model.columns if not column.description
    ]
    if invalid_column_names:
        max_length = 60
        message = f"Columns lack a description: {', '.join(invalid_column_names)}."
        if len(message) > max_length:
            message = f"{message[:max_length]}…"
        return RuleViolation(message=message)


@rule
def has_owner(model: Model) -> RuleViolation | None:
    """A model should have an owner."""
    if not model.meta.get("owner"):
        return RuleViolation(message="Model lacks an owner.")


@rule
def sql_has_reasonable_number_of_lines(
    model: Model, max_lines: int = 200
) -> RuleViolation | None:
    """The SQL query of a model should not be too long."""
    count_lines = len(model.raw_code.split("\n"))
    if count_lines > max_lines:
        return RuleViolation(
            message=f"SQL query too long: {count_lines} lines (> {max_lines})."
        )


@rule(severity=Severity.LOW)
def has_example_sql(model: Model) -> RuleViolation | None:
    """The documentation of a model should have an example query."""
    if model.language == "sql":
        if "```sql" not in (model.description or ""):
            return RuleViolation(
                "The model description does not include an example SQL query."
            )


@rule(rule_filters={is_table()})
def single_pk_defined_at_column_level(model: Model) -> RuleViolation | None:
    """Single-column PK must be defined as a column constraint."""
    for constraint in model.constraints:
        if constraint.type == "primary_key":
            if constraint.columns is not None and len(constraint.columns) == 1:
                return RuleViolation(
                    f"Single-column PK {constraint.columns[0]} must be defined as a "
                    f"column constraint."
                )


@rule(rule_filters={is_table()})
def single_column_uniqueness_at_column_level(model: Model) -> RuleViolation | None:
    """Single-column uniqueness test must be defined as a column test."""
    for data_test in model.tests:
        if data_test.type == "unique_combination_of_columns":
            if len(data_test.kwargs.get("combination_of_columns", [])) == 1:
                return RuleViolation(
                    "Single-column uniqueness test must be defined as a column test."
                )


@rule(rule_filters={is_table()})
def has_uniqueness_test(model: Model) -> RuleViolation | None:
    """Model has uniqueness test for primary key."""
    # ruff: noqa: C901 [too-complex]

    # Single-column PK
    for column in model.columns:
        for column_constraint in column.constraints:
            if column_constraint.type == "primary_key":
                for data_test in column.tests:
                    if data_test.type == "unique":
                        return None
                return RuleViolation(
                    f"No unique constraint defined on PK column {column.name}."
                )

    # Composite PK
    pk_columns: list[str] = []
    for model_constraint in model.constraints:
        if model_constraint.type == "primary_key":
            pk_columns = model_constraint.columns or []
            break

    if not pk_columns:  # No PK, no need for uniqueness test
        return None

    for data_test in model.tests:
        if data_test.type == "unique_combination_of_columns":
            if set(data_test.kwargs.get("combination_of_columns")) == set(pk_columns):  # type: ignore
                return None
    return RuleViolation(
        f"No uniqueness test defined and matching PK {','.join(pk_columns)}."
    )


@rule(rule_filters={is_table()})
def has_no_unused_is_incremental(model: Model) -> RuleViolation | None:
    """Non-incremental model does not make use of is_incremental()."""
    if (
        model.config.get("materialized") != "incremental"
        and "is_incremental()" in model.raw_code
    ):
        return RuleViolation("Non-incremental model makes use of is_incremental().")


@rule
def seed_has_description(seed: Seed) -> RuleViolation | None:
    """A seed should have a description."""
    if not seed.description:
        return RuleViolation(message="Seed lacks a description.")


@rule
def seed_columns_have_description(seed: Seed) -> RuleViolation | None:
    """All columns of a seed should have a description."""
    invalid_column_names = [
        column.name for column in seed.columns if not column.description
    ]
    if invalid_column_names:
        max_length = 60
        message = f"Columns lack a description: {', '.join(invalid_column_names)}."
        if len(message) > max_length:
            message = f"{message[:max_length]}…"
        return RuleViolation(message=message)


@rule
def seed_has_owner(seed: Seed) -> RuleViolation | None:
    """A seed should have an owner."""
    meta = seed.config.get("meta", {})
    if not meta.get("owner"):
        return RuleViolation(message="Seed lacks an owner.")



"""Test rule filters."""
import pytest
from dbt_score.models import Model, Source
from dbt_score.rule_filter import RuleFilter, rule_filter


def test_basic_filter(model1, model2):
    """Test basic filter testing for a specific model."""

    @rule_filter
    def only_model1(model: Model) -> bool:
        """Some description."""
        return model.name == "model1"

    instance = only_model1()  # since the decorator returns a Type
    assert instance.description == "Some description."
    assert instance.evaluate(model1)
    assert not instance.evaluate(model2)


def test_basic_filter_with_sources(source1, source2):
    """Test basic filter testing for a specific source."""

    @rule_filter
    def only_source1(source: Source) -> bool:
        """Some description."""
        return source.name == "table1"

    instance = only_source1()  # since the decorator returns a Type
    assert instance.description == "Some description."
    assert instance.evaluate(source1)
    assert not instance.evaluate(source2)


def test_class_filter(model1, model2):
    """Test basic filter using class."""

    class OnlyModel1(RuleFilter):
        description = "Some description."

        def evaluate(self, model: Model) -> bool:  # type: ignore[override]
            return model.name == "model1"

    instance = OnlyModel1()
    assert instance.description == "Some description."
    assert instance.evaluate(model1)
    assert not instance.evaluate(model2)


def test_class_filter_with_sources(source1, source2):
    """Test basic filter using class."""

    class OnlySource1(RuleFilter):
        description = "Some description."

        def evaluate(self, source: Source) -> bool:  # type: ignore[override]
            return source.name == "table1"

    instance = OnlySource1()
    assert instance.description == "Some description."
    assert instance.evaluate(source1)
    assert not instance.evaluate(source2)


def test_missing_description_rule_filter():
    """Test missing description in filter decorator."""
    with pytest.raises(AttributeError):

        @rule_filter()
        def example_filter(model: Model) -> bool:
            return True


def test_missing_description_rule_class():
    """Test missing description in filter class."""
    with pytest.raises(AttributeError):

        class BadFilter(RuleFilter):
            """Bad example filter."""

            def evaluate(self, model: Model) -> bool:  # type: ignore[override]
                """Evaluate filter."""
                return True


def test_missing_evaluate_rule_class(model1):
    """Test missing evaluate implementation in filter class."""
    with pytest.raises(TypeError):

        class BadFilter(RuleFilter):
            """Bad example filter."""

            description = "Description of the rule."



"""Module computing scores."""

from __future__ import annotations

import math
import typing
from dataclasses import dataclass

from dbt_score.config import Config

if typing.TYPE_CHECKING:
    from dbt_score.evaluation import EvaluableResultsType
from dbt_score.rule import RuleViolation, Severity


@dataclass
class Score:
    """Class representing a score."""

    value: float
    badge: str

    @property
    def rounded_value(self) -> float:
        """Auto-round score down to 1 decimal place."""
        return math.floor(self.value * 10) / 10


class Scorer:
    """Logic for computing scores."""

    # This magic number comes from rule severity.
    # Assuming a rule violation:
    # - A low severity yields a score 2/3
    # - A medium severity yields a score 1/3
    # - A high severity yields a score 0/3
    score_cardinality = 3

    min_score = 0.0
    max_score = 10.0

    def __init__(self, config: Config) -> None:
        """Create a Scorer object."""
        self._config = config

    def score_evaluable(self, evaluable_results: EvaluableResultsType) -> Score:
        """Compute the score of a given evaluable."""
        rule_count = len(evaluable_results)

        if rule_count == 0:
            # No rule? No problem
            score = self.max_score
        elif any(
            rule.severity == Severity.CRITICAL and isinstance(result, RuleViolation)
            for rule, result in evaluable_results.items()
        ):
            # If there's a CRITICAL violation, the score is 0
            score = self.min_score
        else:
            # Otherwise, the score is the weighted average (by severity) of the results
            score = (
                sum(
                    [
                        # The more severe the violation, the more points are lost
                        self.score_cardinality - rule.severity.value
                        if isinstance(result, RuleViolation)  # Either 0/3, 1/3 or 2/3
                        else self.score_cardinality  # 3/3
                        for rule, result in evaluable_results.items()
                    ]
                )
                / (self.score_cardinality * rule_count)
                * self.max_score
            )

        return Score(score, self._badge(score))

    def score_aggregate_evaluables(self, scores: list[Score]) -> Score:
        """Compute the score of a list of evaluables."""
        actual_scores = [s.value for s in scores]
        if 0.0 in actual_scores:
            # Any evaluable with a CRITICAL violation makes the project score 0
            score = Score(self.min_score, self._badge(self.min_score))
        elif len(actual_scores) == 0:
            score = Score(self.max_score, self._badge(self.max_score))
        else:
            average_score = sum(actual_scores) / len(actual_scores)
            score = Score(average_score, self._badge(average_score))
        return score

    def _badge(self, score: float) -> str:
        """Compute the badge of a given score."""
        if score >= self._config.badge_config.first.threshold:
            return self._config.badge_config.first.icon
        elif score >= self._config.badge_config.second.threshold:
            return self._config.badge_config.second.icon
        elif score >= self._config.badge_config.third.threshold:
            return self._config.badge_config.third.icon
        else:
            return self._config.badge_config.wip.icon



"""JSON formatter.

Shape of the JSON output:

```json
{
    "evaluables": {
        "model_foo": {
            "score": 5.0,
            "badge": "🥈",
            "pass": true,
            "results": {
                "rule1": {
                    "result": "OK",
                    "severity": null
                    "message": null
                },
                "rule2": {
                    "result": "WARN",
                    "severity": "medium",
                    "message": "Model lacks a description."
                }
            },
            "type": "model"
        },
        "model_bar": {
            "score": 0.0,
            "badge": "🥉",
            "pass": false,
            "results": {
                "rule1": {
                    "result": "ERR",
                    "message": "Exception message"
                }
            },
            "type": "model"
        },
        "source_baz": {
            "score": 10.0,
            "badge": "🥇",
            "pass": false,
            "results": {
                "rule1": {
                    "result": "ERR",
                    "message": "Exception message"
                }
            },
            "type": "source"
        }
    },
    "project": {
        "score": 5.0,
        "badge": "🥈",
        "pass": false
    }
}
```
"""

import json
from typing import Any

from dbt_score.evaluation import EvaluableResultsType
from dbt_score.formatters import Formatter
from dbt_score.models import Evaluable
from dbt_score.rule import RuleViolation
from dbt_score.scoring import Score


class JSONFormatter(Formatter):
    """Formatter for JSON output."""

    def __init__(self, *args: Any, **kwargs: Any):
        """Instantiate formatter."""
        super().__init__(*args, **kwargs)
        self.evaluable_results: dict[str, dict[str, Any]] = {}
        self._project_results: dict[str, Any]

    def evaluable_evaluated(
        self, evaluable: Evaluable, results: EvaluableResultsType, score: Score
    ) -> None:
        """Callback when an evaluable item has been evaluated."""
        self.evaluable_results[evaluable.name] = {
            "score": score.value,
            "badge": score.badge,
            "pass": score.value >= self._config.fail_any_item_under,
            "results": {},
            "type": type(evaluable).__name__.lower(),
        }
        for rule, result in results.items():
            severity = rule.severity.name.lower()
            if result is None:
                self.evaluable_results[evaluable.name]["results"][rule.source()] = {
                    "result": "OK",
                    "severity": severity,
                    "message": None,
                }
            elif isinstance(result, RuleViolation):
                self.evaluable_results[evaluable.name]["results"][rule.source()] = {
                    "result": "WARN",
                    "severity": severity,
                    "message": result.message,
                }
            else:
                self.evaluable_results[evaluable.name]["results"][rule.source()] = {
                    "result": "ERR",
                    "severity": severity,
                    "message": str(result),
                }

    def project_evaluated(self, score: Score) -> None:
        """Callback when a project has been evaluated."""
        self._project_results = {
            "score": score.value,
            "badge": score.badge,
            "pass": score.value >= self._config.fail_project_under,
        }
        document = {
            "evaluables": self.evaluable_results,
            "project": self._project_results,
        }
        print(json.dumps(document, indent=2, ensure_ascii=False))
